# Copyright 2015 Hewlett-Packard Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from elasticsearch import exceptions as es_exc
from elasticsearch import helpers
import logging
from oslo_utils import fnmatch

from searchlight.common import utils
from searchlight.elasticsearch import ROLE_USER_FIELD

# Refer to ROLE_USER_FIELD
ADMIN_ID_SUFFIX = "_ADMIN"
USER_ID_SUFFIX = "_USER"
VERSION_CONFLICT_EXCEPTION = "version_conflict_engine_exception"


def strip_role_suffix(strip_from, suffix=None):
    if suffix:
        if strip_from.endswith(suffix):
            return strip_from[:-len(suffix)]
    else:
        if strip_from.endswith(ADMIN_ID_SUFFIX):
            return strip_from[:-len(ADMIN_ID_SUFFIX)]
        elif strip_from.endswith(USER_ID_SUFFIX):
            return strip_from[:-len(USER_ID_SUFFIX)]
    return strip_from


def get_metafield(doc, metafield):
    """For fields like _parent and _routing es1.x requires you to request them
    in 'fields', and they're returned as such. ES2.x always returns them but
    at the root document level (like _id). This function copes with both.
    """
    return doc.get(metafield, doc.get('fields', {}).get(metafield))


# String from Alias Failure Exception. See _is_multiple_alias_exception().
ALIAS_EXCEPTION_STRING = "has more than one indices associated with it"

DOC_VALUE_TYPES = ('long', 'integer', 'short', 'boolean', 'byte',
                   'double', 'float', 'date', 'binary', 'geo_point', 'ip')

LOG = logging.getLogger(__name__)


class IndexingHelper(object):
    """Class to abstract away some of the details of indexing documents
    including versioning, parent-child links and separation by role.

    Role separation is reasonably simple; documents with admin-only fields
    will be indexed twice, once for admins and once for normal users, with some
    fields excluded for the second case. The document IDs use the suffixes
    defined in ADMIN_ID_SUFFIX and USER_ID_SUFFIX.

    In the case of parent child relationships things are more complicated. The
    ids in that case follow these rules:
    * parent is role separated
      * child is role separated: the _USER and _ADMIN docs each refer to a
            parent with the same suffix
      * child is NOT separated: this case is not encouraged, but where it
            exists the child documents will use the _USER parent document as
            their _parent to avoid security issues with has_parent query
            fishing
    * parent is not role separated
      * child is role separated: both child docs will use the parent id
            without any prefix
      * child is not role separated: the simple case; no suffix in either
            case
    """

    def __init__(self, plugin):
        self.plugin = plugin

        # Retain some state here for convenience
        self.engine = self.plugin.engine
        self.alias_name = plugin.alias_name_listener
        self.document_type = plugin.document_type

    @property
    def index_chunk_size(self):
        return 200

    def _index_alias_multiple_indexes_bulk(self, documents=None, actions=None,
                                           versions=None):
        """A bulk operation failed by trying to access an alias that has
           multiple indexes. To rememdy this we will need to iterate on all
           indexes within the alias and retry the bulk operation individually.
        """
        indexes = self.engine.indices.get_alias(index=self.alias_name)
        for index_name in indexes:
            try:
                if documents:
                    result = helpers.bulk(
                        client=self.engine,
                        index=index_name,
                        doc_type=self.document_type,
                        chunk_size=self.index_chunk_size,
                        actions=self._prepare_actions(documents,
                                                      versions))
                if actions:
                    result = helpers.bulk(
                        client=self.engine,
                        index=index_name,
                        doc_type=self.document_type,
                        chunk_size=self.index_chunk_size,
                        actions=actions)
                LOG.debug("Indexing result: %s", result)
            except Exception as e:
                # Log the error and continue to the next index.
                format_msg = {
                    'doc': self.document_type,
                    'msg': str(e)
                }
                LOG.error("Failed Indexing %(doc)s: %(msg)s" % format_msg)

    def _index_alias_multiple_indexes_get(self, doc_id, routing=None):
        """Getting a document from an alias with multiple indexes will fail.
           We need to retrieve it from one of the indexes. We will choose
           the latest index. Since the indexes are named with a timestamp,
           a reverse sort will bring the latest to the front.
        """
        indexes = self.engine.indices.get_alias(index=self.alias_name)
        index_list = indexes.keys()
        index_list.sort(reverse=True)
        try:
            if routing:
                return self.engine.get(
                    index=index_list[0],
                    doc_type=self.document_type,
                    id=doc_id,
                    routing=routing
                )
            else:
                return self.engine.get(
                    index=index_list[0],
                    doc_type=self.document_type,
                    id=doc_id
                )
        except Exception as e:
            format_msg = {
                'doc': self.document_type,
                'msg': str(e)
            }
            LOG.error("Failed Indexing %(doc)s: %(msg)s" % format_msg)

    def save_document(self, document, version=None):
        if version:
            self.save_documents([document], [version])
        else:
            self.save_documents([document])

    def save_documents(self, documents, versions=None, index=None):
        """Send list of serialized documents into search engine.

           Warning: Index vs Alias usage.
           Listeners [plugins/*/notification_handlers.py]:
           When the plugin listeners are indexing documents, we will want
           to use the normal ES alias for their resource group. In this case
           the index parameter will not be set. Listeners are by far the most
           common usage case.

           Re-Indexing [plugins/base.py::index_initial_data()]:
           When we are re-indexing we will want to use the new ES index.
           Bypassing the alias means we will not send duplicate documents
           to the old index. In this case the index will be set. Re-indexing
           is an event that will rarely happen.
        """
        if not index:
            use_index = self.alias_name
        else:
            use_index = index

        for document in documents:
            # NOTE: In Elasticsearch 2.0 field names cannot contain '.', change
            # '.' to '__'.
            utils.replace_dots_in_field_names(document)

        try:
            result = helpers.bulk(
                client=self.engine,
                index=use_index,
                doc_type=self.document_type,
                chunk_size=self.index_chunk_size,
                actions=self._prepare_actions(documents, versions))
        except helpers.BulkIndexError as e:
            err_msg = []
            for err in e.errors:
                if (err['index']['error']['type'] !=
                    VERSION_CONFLICT_EXCEPTION and
                        err['index']['status'] != 409):
                    raise
                err_msg.append("id %(_id)s: %(error)s" % err['index'])
            LOG.warning('Version conflict %s' % ';'.join(err_msg))
            result = 0
        except es_exc.RequestError as e:
            if _is_multiple_alias_exception(e):
                LOG.error("Alias [%(a)s] with multiple indexes error" %
                          {'a': self.alias_name})
                self._index_alias_multiple_indexes_bulk(documents=documents,
                                                        versions=versions)

            result = 0
        LOG.debug("Indexing result: %s", result)

    def delete_document(self, document):
        """'document' must contain an '_id', but can include '_parent',
        '_version' and '_routing', each of which will be passed to
        the bulk helper
        """
        self.delete_documents([document])

    def delete_documents(self, metadocs, override_role_separation=False):
        """Each metadoc should be a dict with at an _id, and if
         applicable, a _parent. override_role_separation will treat the _ids
         and _parents in the documents as their actual indexed values
         rather than determining role separation
         """
        def _get_delete_action(doc, id_suffix=''):
            action = {'_op_type': 'delete', '_id': doc['_id'] + id_suffix}

            if doc.get('_version'):
                action['_version'] = doc['_version']
                action['_version_type'] = 'external'

            parent_entity_id = doc.get('_parent')
            if parent_entity_id:
                if (not override_role_separation and
                        self.plugin.parent_plugin.requires_role_separation):
                    # Default to _USER; defaulting to _ADMIN causes a
                    # security issue because of potential fishing queries
                    parent_entity_id += (id_suffix or USER_ID_SUFFIX)
                action['_parent'] = parent_entity_id
            if '_routing' in doc:
                action['_routing'] = doc['_routing']
            return action

        actions = []
        for metadoc in metadocs:
            if (not override_role_separation and
                    self.plugin.requires_role_separation):
                actions.extend([
                    _get_delete_action(metadoc, ADMIN_ID_SUFFIX),
                    _get_delete_action(metadoc, USER_ID_SUFFIX)])
            else:
                actions.append(_get_delete_action(metadoc))

        try:
            helpers.bulk(
                client=self.plugin.engine,
                index=self.alias_name,
                doc_type=self.document_type,
                actions=actions
            )
        except helpers.BulkIndexError as exc:
            exc_payload = exc.errors
            doc_ids = ', '.join(e['delete']['_id'] for e in exc_payload)

            if all(e['delete']['status'] == 404 for e in exc_payload):
                LOG.warning(
                    "Error deleting %(doc_type)s %(ids)s; "
                    "already deleted" %
                    {"doc_type": self.plugin.document_type, "ids": doc_ids})

            elif all(e['delete']['status'] == 409 for e in exc_payload):
                # This *should* never happen. If it does, something has gone
                # wrong but leaving this here for now
                LOG.warning(
                    "Error deleting %(doc_type)s %(ids)s; newer versions "
                    "of some documents have been indexed" %
                    {"doc_type": self.plugin.document_type, "ids": doc_ids})
            else:
                raise
        except es_exc.RequestError as e:
            if _is_multiple_alias_exception(e):
                LOG.error("Alias [%(a)s] with multiple indexes error" %
                          {'a': self.alias_name})
                self._index_alias_multiple_indexes_bulk(actions=actions)

    def delete_documents_with_parent(self, parent_id, version=None):
        # This is equivalent in result to _parent: parent_id but offers
        # a significant performance boost because of the implementation
        # of _parent filtering
        parent_type = self.plugin.parent_plugin_type()

        # It's easier to retrieve the actual parent id here because otherwise
        # we have to figure out role separation. _parent is (in 1.x) not
        # return by default and has to be requested in 'fields'; in 2.x it
        # IS returned by default (but not in fields, yay)
        query = {
            'fields': ['_parent', '_routing']
        }

        if (self.plugin.parent_plugin and
                self.plugin.parent_plugin.requires_role_separation):
            # There will be documents with the _USER suffix; there may also
            # be those with _ADMIN suffixes if this plugin type is separated
            full_parent_ids = [
                '%s#%s%s' % (parent_type, parent_id, ADMIN_ID_SUFFIX),
                '%s#%s%s' % (parent_type, parent_id, USER_ID_SUFFIX)
            ]
            query['query'] = {'terms': {'_parent': full_parent_ids}}
        else:
            full_parent_id = '%s#%s' % (parent_type, parent_id)
            query['query'] = {'term': {'_parent': full_parent_id}}

        documents = helpers.scan(
            client=self.engine,
            index=self.alias_name,
            doc_type=self.document_type,
            query=query)

        def _get_delete_action(doc):
            routing = get_metafield(doc, '_routing')
            delete_action = {'_id': doc['_id'],
                             '_parent': get_metafield(doc, '_parent')}
            if routing:
                delete_action['_routing'] = routing
            return delete_action

        to_delete = [_get_delete_action(d) for d in documents]

        # Use the parent version tag; we're instructing elasticsearch to mark
        # all the deleted child documents as 'don't apply updates received
        # after 'version' so the fact that they don't match the child
        # 'updated_at' fields is irrelevant
        if version:
            for action in to_delete:
                action['_version'] = version

        self.delete_documents(to_delete, override_role_separation=True)
        return to_delete

    def delete_document_unknown_parent(self, doc_id, version=None):
        """Deletes a document that requires routing but where the routing is
        not known. Since this involves a query to find it, there is a potential
        race condition in between storing and indexing the document. A search
        is run to find the routing parent, and the document is then deleted.
        """
        search_doc_id = doc_id
        if self.plugin.requires_role_separation:
            # Just look for the admin document though it doesn't matter which.
            # This has to match whatever we strip later with strip_role_suffix
            search_doc_id += ADMIN_ID_SUFFIX

        # Necessary to request 'fields' for e-s 1.x, not for 2.x
        query = {'filter': {'term': {'_id': search_doc_id}},
                 'fields': ['_parent', '_routing']}
        search_results = self.engine.search(index=self.alias_name,
                                            doc_type=self.document_type,
                                            body=query)
        total_hits = search_results['hits']['total']
        if not total_hits:
            ctx = {'doc_type': self.document_type, 'id': doc_id}
            LOG.warning(
                "No results found for %(doc_type)s id %(id)s; can't find "
                "routing to delete" % ctx)
            return

        # There are results. Check that there's only one unique result (may be
        # two in different indices from the same alias). ES 1.x and 2.x differ
        # slightly; metafields are returned by default and not in 'fields' in
        # 2.x whether you like it or not
        distinct_results = set((r['_id'], get_metafield(r, '_parent'))
                               for r in search_results['hits']['hits'])
        if len(distinct_results) != 1:
            ctx = {'count': len(distinct_results),
                   'results': ", ".join(distinct_results),
                   'doc_type': self.document_type,
                   'id': doc_id}
            LOG.error("%(count)d distinct results (%(results)s) found for "
                      "get_document_by_query for %(doc_type)s id %(id)s" %
                      ctx)

        first_hit = search_results['hits']['hits'][0]
        routing = get_metafield(first_hit, '_routing')
        parent = get_metafield(first_hit, '_parent')
        parent = strip_role_suffix(parent, ADMIN_ID_SUFFIX)

        LOG.debug("Deleting %s id %s with parent %s routing %s",
                  self.document_type, doc_id, parent, routing)
        delete_info = {'_id': doc_id, '_parent': parent}
        if routing:
            delete_info['_routing'] = routing
        if version:
            delete_info['_version'] = version
        self.delete_document(delete_info)

    def get_document(self, doc_id, for_admin=False, routing=None):
        if self.plugin.requires_role_separation:
            doc_id += (ADMIN_ID_SUFFIX if for_admin else USER_ID_SUFFIX)

        try:
            if routing:
                return self.engine.get(
                    index=self.alias_name,
                    doc_type=self.document_type,
                    id=doc_id,
                    routing=routing
                )
            else:
                return self.engine.get(
                    index=self.alias_name,
                    doc_type=self.document_type,
                    id=doc_id
                )
        except es_exc.RequestError:
            # TODO(ricka) Verify this is the IllegalArgument exception.
            LOG.error("Alias [%(alias)s] with multiple indexes error" %
                      {'alias': self.alias_name})
            #
            return self._index_alias_multiple_indexes_get(
                doc_id=doc_id, routing=routing)

    def get_docs_by_nested_field(self, path, field, value, version=False):
        """Query ElasticSearch based on a nested field. The caller will
           need to specify the path of the nested field as well as the
           field itself. We will include the 'version' field if commanded
           as such by the caller.
        """
        # Set up query for accessing a nested field.
        nested_field = path + "." + field
        body = {"query": {"nested": {
                "path": path, "query": {"term": {nested_field: value}}}}}
        if version:
            body['version'] = True
        try:
            return self.engine.search(index=self.alias_name,
                                      doc_type=self.document_type,
                                      body=body, ignore_unavailable=True)
        except Exception as exc:
            LOG.warning(
                'Error querying %(p)s %(f)s. Error %(exc)s' %
                {'p': path, 'f': field, 'exc': exc})
            raise

    def update_document(self, document, doc_id, update_as_script,
                        expected_version=None):
        """Updates are a little simpler than inserts because the documents
        already exist. Note that scripted updates are not filtered in the same
        way as partial document updates. Script updates should be passed as
        a dict {"script": .., "parameters": ..}. Partial document updates
        should be the raw document.
        """
        def _get_update_action(source, id_suffix=''):
            action = {'_id': doc_id + id_suffix, '_op_type': 'update'}
            if expected_version:
                action['_version'] = expected_version
            if update_as_script:
                action.update(source)
            else:
                action['doc'] = source

            routing_field = self.plugin.routing_field
            if routing_field:
                action['_routing'] = source[routing_field]

            return action

        if self.plugin.requires_role_separation:
            user_doc = (self._remove_admin_fields(document)
                        if update_as_script else document)
            actions = [_get_update_action(document, ADMIN_ID_SUFFIX),
                       _get_update_action(user_doc, USER_ID_SUFFIX)]
        else:
            actions = [_get_update_action(document)]
        try:
            result = helpers.bulk(
                client=self.engine,
                index=self.alias_name,
                doc_type=self.document_type,
                chunk_size=self.index_chunk_size,
                actions=actions)
            LOG.debug("Update result: %s", result)
        except es_exc.RequestError as e:
            if _is_multiple_alias_exception(e):
                LOG.error("Alias [%(a)s] with multiple indexes error" %
                          {'a': self.alias_name})
                self._index_alias_multiple_indexes_bulk(actions=actions)

    def _prepare_actions(self, documents, versions=None):
        """Returns a generator of indexable 'actions'. If the plugin requires
        role-based separation, two actions will be yielded per document,
        otherwise one. _id and USER_ROLE_FIELD are set as appropriate
        """
        def _get_index_action(source, version=None, id_suffix=''):
            """Return an 'action' the ES bulk indexer understands"""
            action = {
                '_id': source[self.plugin.get_document_id_field()] + id_suffix,
                '_source': source,
                '_op_type': 'index'
            }
            if version:
                action['_version_type'] = 'external'
                action['_version'] = version

            parent_field = self.plugin.get_parent_id_field()
            routing_field = self.plugin.routing_field
            if parent_field:
                parent_id = source[parent_field]
                if self.plugin.parent_plugin.requires_role_separation:
                    # Default to _USER; defaulting to _ADMIN causes a
                    # security issue because of potential fishing queries
                    parent_id += (id_suffix or USER_ID_SUFFIX)
                action['_parent'] = parent_id
            if routing_field:
                action['_routing'] = source[routing_field]
            return action

        for index, document in enumerate(documents):
            # Although elasticsearch copies the input when indexing, it's
            # easier from a debugging and testing perspective not to meddle
            # with the input, so make a copy
            document = copy.deepcopy(document)
            version = versions[index] if versions else None

            if self.plugin.include_region_name:
                document['region_name'] = self.plugin.region_name

            if self.plugin.requires_role_separation:
                LOG.debug("Applying role separation to %s id %s" %
                          (self.plugin.name,
                           document[self.plugin.get_document_id_field()]))
                document[ROLE_USER_FIELD] = 'admin'
                yield _get_index_action(document,
                                        version=version,
                                        id_suffix=ADMIN_ID_SUFFIX)

                user_doc = self._remove_admin_fields(document)
                user_doc[ROLE_USER_FIELD] = 'user'
                yield _get_index_action(user_doc,
                                        version=version,
                                        id_suffix=USER_ID_SUFFIX)

            else:
                LOG.debug("Not applying role separation to %s id %s" %
                          (self.plugin.name,
                           document[self.plugin.get_document_id_field()]))
                document[ROLE_USER_FIELD] = ['user', 'admin']
                yield _get_index_action(document, version=version)

    def _remove_admin_fields(self, document):
        """Prior to indexing, remove any fields that shouldn't be indexed
        and available to users who do not have administrative privileges.
        Returns a copy of the document even if there's nothing to remove.
        """
        sanitized_document = {}
        for k, v in document.items():
            # Only return a field if it doesn't have ANY matches against
            # admin_only_fields
            if not any(fnmatch.fnmatch(k, field)
                       for field in self.plugin.admin_only_fields):
                sanitized_document[k] = v

        return sanitized_document

    @classmethod
    def apply_doc_values(cls, mapping):
        """Sets 'doc_values' on fields in a mapping which allows queries to be
        run on fields directly off disk, saving memory in analysis operations.
        Currently all fields with the exception of analyzed strings can be set
        as doc_values. Elasticsearch 2.x will make doc_values the default.
        """
        def apply_doc_values(field_def):
            if field_def.get('type', 'object') in ('nested', 'object'):
                for _, nested_def in field_def['properties'].items():
                    apply_doc_values(nested_def)
            else:
                if 'doc_values' not in field_def:
                    if field_def['type'] in DOC_VALUE_TYPES:
                        field_def['doc_values'] = True
                    elif (field_def['type'] == 'string' and
                          field_def.get('index', '') == 'not_analyzed'):
                        field_def['doc_values'] = True

                for _, multidef in field_def.get('fields', {}).items():
                    apply_doc_values(multidef)

        # Check dynamic templates. These are a list of dicts each with a single
        # key (the template name) and a mapping definition
        for dynamic_template in mapping.get('dynamic_templates', []):
            for dyn_field, dyn_mapping in dynamic_template.items():
                apply_doc_values(dyn_mapping['mapping'])

        for field, definition in mapping['properties'].items():
            apply_doc_values(definition)

    def simple_search(self, type, query):
        search_results = self.engine.search(index=self.alias_name,
                                            doc_type=type,
                                            body=query)
        return search_results['hits']


def _is_multiple_alias_exception(e):
    """Verify that this exception is specifically the IllegalArgument
       exception when there are multiple indexes for an alias. There
       is no clean way of verifying this is the case. There are multiple
       ES RequestError exceptions that return a 400 IllegalArgument.
       In this particular case, we are expecting a message in the
       exception like this:

           ElasticsearchIllegalArgumentException[Alias [alias] has more
           than one indices associated with it [[idx1, idx2]], can't
           execute a single index op]

       We will be dirty and parse the exception message. We need to
       check the validity of ALIAS_EXCEPTION_STRING in future
       ElasticSearch versions.
    """
    if ALIAS_EXCEPTION_STRING in getattr(e, 'error', ''):
        # Elasticsearch 1.x
        return True

    # ES 2 - the error info's in e.info.error.reason
    exc_info_error = getattr(e, 'info', {}).get('error', {})
    if ALIAS_EXCEPTION_STRING in exc_info_error.get('reason', ''):
        return True

    return False



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.cinder import serialize_cinder_volume
from searchlight.elasticsearch.plugins.cinder \
    import volumes_notification_handler
from searchlight.elasticsearch.plugins import openstack_clients

LOG = logging.getLogger(__name__)


# TODO(sjmc7): Parameterize once we have plugin configs
LIST_LIMIT = 100


class VolumeIndex(base.IndexBase):
    NotificationHandlerCls = volumes_notification_handler.VolumeHandler

    # Will be combined with 'admin_only_fields' from config
    ADMIN_ONLY_FIELDS = ['os-vol-mig-status-attr:*', 'os-vol-host-attr:*',
                         'migration_status']

    @classmethod
    def get_document_type(self):
        return resource_types.CINDER_VOLUME

    def get_mapping(self):
        str_not_analyzed = {'type': 'string', 'index': 'not_analyzed'}
        return {
            'dynamic': True,
            'properties': {
                'attachments': {
                    'type': 'nested',
                    'properties': {
                        'attachment_id': str_not_analyzed,
                        'device': str_not_analyzed,
                        'host_name': str_not_analyzed,
                        'id': str_not_analyzed,
                        'server_id': str_not_analyzed,
                        'attached_at': {'type': 'date'}
                    }
                },
                'availability_zone': {'type': 'string',
                                      'index': 'not_analyzed'},
                'consistencygroup_id': {
                    'type': 'string', 'index': 'not_analyzed'
                },
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'bootable': {'type': 'boolean'},
                'created_at': {'type': 'date'},
                'description': {'type': 'string'},
                'encrypted': {'type': 'boolean'},
                'migration_status': {
                    'type': 'string', 'index': 'not_analyzed'},
                'multiattach': {'type': 'boolean'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'os-vol-host-attr:host': {
                    'type': 'string', 'index': 'not_analyzed'
                },
                'os-vol-mig-status-attr:migstat': {
                    'type': 'string', 'index': 'not_analyzed'
                },
                'os-vol-mig-status-attr:name_id': {
                    'type': 'string', 'index': 'not_analyzed'
                },
                'os-vol-tenant-attr:tenant_id': {
                    'type': 'string', 'index': 'not_analyzed'
                },
                'os-volume-replication:driver_data': {
                    'type': 'string', 'index': 'not_analyzed'
                },
                'os-volume-replication:extended_status': {
                    'type': 'string', 'index': 'not_analyzed'
                },
                'project_id': {'type': 'string', 'index': 'not_analyzed'},
                'replication_status': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                # Using long here to consistent with glance.images
                'size': {'type': 'long'},
                'snapshot_id': {'type': 'string', 'index': 'not_analyzed'},
                'source_volid': {'type': 'string', 'index': 'not_analyzed'},
                'status': {'type': 'string', 'index': 'not_analyzed'},
                'tenant_id': {'type': 'string', 'index': 'not_analyzed'},
                'volume_type': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    }
                },
                'updated_at': {'type': 'date'},
                'user_id': {'type': 'string', 'index': 'not_analyzed'},
            },
            "_meta": {
                "snapshot_id": {
                    "resource_type": resource_types.CINDER_SNAPSHOT
                },
                "tenant_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "attachments.server_id": {
                    "resource_type": resource_types.NOVA_SERVER
                },
                "os-vol-tenant-attr:tenant_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "user_id": {
                    "resource_type": resource_types.KEYSTONE_USER
                }
            }
        }

    @property
    def admin_only_fields(self):
        from_conf = super(VolumeIndex, self).admin_only_fields
        return VolumeIndex.ADMIN_ONLY_FIELDS + from_conf

    @property
    def facets_with_options(self):
        return ('status', 'availability_zone', 'bootable', 'volume_type',
                'encrypted', 'migration_status', 'replication_status',
                'multiattach')

    @property
    def facets_excluded(self):
        """A map of {name: allow_admin} that indicate which
        fields should not be offered as facet options, or those that should
        only be available to administrators.
        """
        return {'os-vol-tenant-attr:tenant_id': True,
                'tenant_id': True,
                'migration_status': True,
                'os-vol-host-attr:host': True,
                'replication_status': True,
                'user_id': True,
                'os-vol-mig-status-attr:migstat': True,
                'os-vol-mig-status-attr:name_id': True,
                'os-volume-replication:driver_data': True,
                'os-volume-replication:extended_status': True,
                'source_volid': True,
                'migration_status': True}

    @property
    def resource_allowed_policy_target(self):
        return 'volume:get_all'

    @property
    def service_type(self):
        return 'volume'

    def _get_rbac_field_filters(self, request_context):
        """Return any RBAC field filters to be injected into an indices
        query. Document type will be added to this list.
        """
        return [
            {'term': {'project_id': request_context.owner}}
        ]

    def get_objects(self):
        """Generator that lists all cinder volumes owned by all tenants."""
        LOG.debug("Cinder volumes get_objects started")
        has_more = True
        marker = None
        while has_more:
            volumes = openstack_clients.get_cinderclient().volumes.list(
                limit=LIST_LIMIT,
                search_opts={'all_tenants': True},
                marker=marker
            )

            if not volumes:
                # Definitely no more; break straight away
                break

            # volumes.list always returns a list so we can grab the last id
            has_more = len(volumes) == LIST_LIMIT
            marker = volumes[-1].id

            for volume in volumes:
                yield volume

    def serialize(self, volume):
        return serialize_cinder_volume(volume)



# Copyright (c) 2011 OpenStack Foundation
# Copyright 2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Searchlight"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import policy

from searchlight.common import exception
from searchlight.common import policies
from searchlight import service_policies


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

DEFAULT_RULES = policy.Rules.from_dict({
    'context_is_admin': 'role:admin',
    'default': '@',
})


class Enforcer(policy.Enforcer):
    """Responsible for loading and enforcing rules"""

    def __init__(self):
        if CONF.find_file(CONF.oslo_policy.policy_file):
            kwargs = {'rules': None, 'use_conf': True}
        else:
            kwargs = {'rules': DEFAULT_RULES, 'use_conf': False}
        super(Enforcer, self).__init__(CONF, overwrite=False, **kwargs)
        self.register_defaults(policies.list_rules())

    def add_rules(self, rules):
        """Add new rules to the Rules object"""
        self.set_rules(rules, overwrite=False, use_conf=self.use_conf)

    def enforce(self, context, action, target):
        """Verifies that the action is valid on the target in this context.

           :param context: Searchlight request context
           :param action: String representing the action to be checked
           :param target: Dictionary representing the object of the action.
           :raises: `searchlight.common.exception.Forbidden`
           :returns: A non-False value if access is allowed.
        """
        credentials = context.to_dict()
        return super(Enforcer, self).enforce(action, target, credentials,
                                             do_raise=True,
                                             exc=exception.Forbidden,
                                             action=action)

    def check(self, context, action, target):
        """Verifies that the action is valid on the target in this context.

           :param context: Searchlight request context
           :param action: String representing the action to be checked
           :param target: Dictionary representing the object of the action.
           :returns: A non-False value if access is allowed.
        """
        credentials = context.to_dict()
        return super(Enforcer, self).enforce(action, target, credentials)

    def check_is_admin(self, context):
        """Check if the given context is associated with an admin role,
           as defined via the 'context_is_admin' RBAC rule.

           :param context: Searchlight request context
           :returns: A non-False value if context role is admin.
        """
        return self.check(context, 'context_is_admin', context.to_dict())


class CatalogSearchRepoProxy(object):

    def __init__(self, search_repo, context, search_policy):
        self.context = context
        self.policy = search_policy
        self.search_repo = search_repo

    def search(self, *args, **kwargs):
        self.policy.enforce(self.context, 'search:query',
                            self.context.policy_target)
        return self.search_repo.search(*args, **kwargs)

    def plugins_info(self, *args, **kwargs):
        self.policy.enforce(self.context, 'search:plugins_info',
                            self.context.policy_target)
        return self.search_repo.plugins_info(*args, **kwargs)

    def facets(self, *args, **kwargs):
        self.policy.enforce(self.context, 'search:facets',
                            self.context.policy_target)
        return self.search_repo.facets(*args, **kwargs)


def plugin_allowed(policy_enforcer, context, plugin):
    """Returns True or False indicating whether a plugin should be available
    for API operations to a given user.
    """
    resource_type = plugin.get_document_type()
    policy_action = 'resource:%s' % resource_type
    service_type = plugin.service_type
    target = context.policy_target
    if not policy_enforcer.check(context, policy_action, target):
        LOG.debug("Policy for '%s' forbids '%s'",
                  policy_action, resource_type)
        return False

    # Now try any related policy action
    service_policy_action = plugin.resource_allowed_policy_target

    service_enforcer = service_policies.get_enforcer_for_service(service_type)

    if service_enforcer and service_policy_action:
        if not resource_allowed(service_enforcer, context,
                                service_policy_action):
            LOG.debug("Policy for '(%s) %s' forbids '%s'",
                      service_type, service_policy_action, resource_type)
            return False

    # (sjmc7) Current decision is not to enforce policy files for services,
    # though that might be an option that could be added. If a service
    # enforcer is configured improperly the server won't start.

    # At this point all tests that exist have passed, so allow the plugin
    return True


def resource_allowed(enforcer, context, resource_allowed_action, target={}):
    """Check whether 'resource_allowed_action' is allowed by 'enforcer' for
    the given context.
    """
    credentials = context.to_policy_values()
    if 'tenant_id' not in credentials:
        credentials['tenant_id'] = credentials.get('project_id', None)
    if 'is_admin' not in credentials:
        credentials['is_admin'] = context.is_admin

    if not target:
        # This allows 'admin_or_owner' type rules to work
        target = {
            'project_id': context.tenant,
            'user_id': context.user,
            'tenant_id': context.tenant
        }
    return enforcer.enforce(resource_allowed_action, target, credentials)



# Copyright (c) 2016 Hewlett-Packard Enterprise Development L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from oslo_config import cfg
from oslo_policy import policy

from searchlight.common import policies


LOG = logging.getLogger(__name__)


policy_file_opts = [
    cfg.StrOpt("service_policy_path",
               default='',
               help="Base path for policy files. If not specified, "
                    "oslo.config search paths will be used"),
    cfg.DictOpt("service_policy_files",
                default={},
                help="Service policy files of the form <service type>:<path>,"
                     "<service_type>:<path>. Service type should be e.g."
                     "'compute', 'volume'")
]
cfg.CONF.register_opts(policy_file_opts, group="service_policies")


_ENFORCERS = None


class MissingPolicyFile(Exception):
    pass


def _get_enforcers():
    global _ENFORCERS
    if not _ENFORCERS:
        _ENFORCERS = {}
        pol_files = cfg.CONF.service_policies.service_policy_files
        for service, pol_file in pol_files.items():
            base_path = str(cfg.CONF.service_policies.service_policy_path)
            service_policy_path = os.path.join(base_path,
                                               pol_file)
            enforcer = policy.Enforcer(cfg.CONF, service_policy_path)
            missing_config_file = False

            # oslo.policy's approach to locating these files seems to be
            # changing; current master doesn't raise an exception
            try:
                enforcer.load_rules()
                enforcer.register_defaults(policies.list_rules())
                if not enforcer.policy_path:
                    missing_config_file = True
            except cfg.ConfigFilesNotFoundError:
                missing_config_file = True

            if missing_config_file:
                LOG.error("Policy file for service %(service)s not found"
                          " in %(policy_file)s (base path %(base)s)" %
                          {"service": service, "policy_file": pol_file,
                           "base": service_policy_path})
                raise MissingPolicyFile(
                    "Could not find policy file %(pol_file)s for service "
                    "type %(service)s" % {'pol_file': pol_file,
                                          'service': service})

            LOG.debug("Adding policy enforcer for %s" % service)
            _ENFORCERS[service] = enforcer

    return _ENFORCERS


def _reset_enforcers():
    global _ENFORCERS
    _ENFORCERS = None


def check_policy_files():
    """Verifies that any configured policy files exist and are valid"""
    _get_enforcers()


def get_enforcer_for_service(service_type):
    enforcers = _get_enforcers()
    if service_type in enforcers:
        return enforcers[service_type]

    LOG.debug("No policy file configured for %s", service_type)
    return None



# Copyright 2015 Hewlett-Packard Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock

from searchlight.elasticsearch.plugins.designate import \
    recordsets as recordsets_plugin
from searchlight.elasticsearch import ROLE_USER_FIELD
from searchlight import pipeline
from searchlight.tests.unit import utils as unit_test_utils
import searchlight.tests.utils as test_utils


now = datetime.datetime.utcnow()
five_mins = datetime.timedelta(minutes=5)
updated_now = now.strftime('%Y-%m-%dT%H:%M:%SZ')
created_now = (now - five_mins).strftime('%Y-%m-%dT%H:%M:%SZ')

USER1 = u'27f4d76b-be62-4e4e-aa33bb11cc55'

TENANT1 = u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0'
TENANT2 = u'c6993374-7c4b-4f18-b317-85e3acdfd259'

# Zone ids
ZONE_ID1 = u'9b4c2de5-dead-0000-beef-e3b99695aa69'
ZONE_ID2 = u'0c3dae67-c4d3-42d6-9d8d-98da56321bca'

ID1 = u'4d6c2de5-c4d3-0000-4e4e-acd99695a222'
ID2 = u'98543543-3443-acdf-deac-3423423cdadf'
ID3 = u'90754354-bc43-234b-12b5-789234bcdefa'


def _recordset_fixture(id, zone_id, tenant_id, name, **kwargs):
    recordset = {
        'tenant_id': tenant_id,
        'zone_id': zone_id,
        'name': name,
        'ttl': 3600,
        'status': 'pending',
        'created_at': created_now,
        'updated_at': updated_now,
        "id": id
    }
    recordset.update(**kwargs)
    return recordset


class TestRecordsetPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestRecordsetPlugin, self).setUp()
        self.plugin = recordsets_plugin.RecordSetIndex()

        self._create_fixtures()

        self.mock_session = mock.Mock()
        self.mock_session.get_endpoint.return_value = \
            'http://localhost/glance/v2'
        patched_ses = mock.patch(
            'searchlight.elasticsearch.plugins.openstack_clients._get_session',
            return_value=self.mock_session)
        patched_ses.start()
        self.addCleanup(patched_ses.stop)

    def _create_fixtures(self):
        self.recordset1 = _recordset_fixture(
            ID1, ZONE_ID1, TENANT1, name='www.test.com.', type='A',
            records=['192.0.2.1'])
        self.recordset2 = _recordset_fixture(
            ID1, ZONE_ID1, TENANT1, name='www2.test.com.', type='A',
            records=['192.0.5.1', '192.0.1.1'])
        self.recordset3 = _recordset_fixture(
            ID1, ZONE_ID2, TENANT1, name='www.other.com.', type='A',
            records=['192.0.128.128'])

        self.recordsets = (self.recordset1, self.recordset2, self.recordset3)

    def test_missing_updated(self):
        """Designate records don't always have a value for 'updated'"""
        set_to_test = dict(self.recordset1)
        set_to_test['updated_at'] = None
        serialized = self.plugin.serialize(set_to_test)
        self.assertEqual(created_now, serialized['updated_at'])

    def test_facets(self):
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=False
        )

        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        mock_engine.search.return_value = {
            'aggregations': {
                'status': {
                    'buckets': [{'key': 'pending', 'doc_count': 2}]
                },
                'type': {
                    'buckets': [{'key': 'A', 'doc_count': 2}]
                }
            },
            'hits': {'total': 2}
        }

        facets, _ = self.plugin.get_facets(fake_request.context)

        zone_id_facet = list(filter(lambda f: f['name'] == 'zone_id',
                                    facets))[0]
        expected_zone_id = {
            'name': 'zone_id',
            'type': 'string',
            'resource_type': self.plugin.parent_plugin_type(),
            'parent': True
        }

        self.assertEqual(expected_zone_id, zone_id_facet)

        status_facet = list(filter(lambda f: f['name'] == 'status', facets))[0]
        expected_status = {
            'name': 'status',
            'options': [{'key': 'pending', 'doc_count': 2}],
            'type': 'string'
        }
        self.assertEqual(expected_status, status_facet)

        expected_agg_query = {
            'aggs': dict(unit_test_utils.simple_facet_field_agg(name)
                         for name in ('status', 'type')),
            'query': {
                'bool': {
                    'filter': {
                        'bool': {
                            'must': {'term': {ROLE_USER_FIELD: 'user'}},
                            'should': [{'term': {'project_id': TENANT1}}],
                            'minimum_should_match': 1
                        }
                    }
                }
            }
        }
        mock_engine.search.assert_called_with(
            index=self.plugin.alias_name_search,
            doc_type=self.plugin.get_document_type(),
            body=expected_agg_query,
            ignore_unavailable=True,
            size=0
        )

    def test_serialize(self):
        serialized = self.plugin.serialize(self.recordset1)
        self.assertEqual(TENANT1, serialized['project_id'])

    def test_create_recordset(self):
        """Test recordset create event"""
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        create_event = (
            'dns.recordset.create',
            {
                "zone_id": "3081593e-10ca-408c-af77-1397e689c177",
                "tenant_id": "80264096ac454d3d904002491fafe2ec",
                "created_at": "2016-03-31T05:48:53.000000",
                "description": None,
                "updated_at": None,
                "records": [{
                    "status": "PENDING",
                    "zone_id": "3081593e-10ca-408c-af77-1397e689c177",
                    "managed": False,
                    "managed_resource_id": None,
                    "managed_resource_type": None,
                    "tenant_id": "80264096ac454d3d904002491fafe2ec",
                    "created_at": "2016-03-31T05:48:54.000000",
                    "managed_extra": None,
                    "updated_at": None,
                    "managed_plugin_type": None,
                    "version": 1,
                    "managed_plugin_name": None,
                    "managed_tenant_id": None,
                    "action": "CREATE",
                    "hash": "8402992213c4c985429e9f2b23a18847",
                    "managed_resource_region": None,
                    "recordset_id": "cfe39618-49da-4877-914b-f25ef0fb3dc1",
                    "data": "10.0.0.1",
                    "id": "01835dc4-b805-4c8f-b962-2a96880f40c7",
                    "serial": 1459403333,
                    "description": None
                }],
                "version": 1,
                "ttl": None,
                "type": "A",
                "id": "cfe39618-49da-4877-914b-f25ef0fb3dc1",
                "name": "www.myzone.net."
            },
            '2016-03-31 05:48:54.102791'
        )
        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        with mock.patch.object(self.plugin.index_helper,
                               'save_document') as mock_save:
            type_handler = event_handlers.get(create_event[0], None)
            result = type_handler(*create_event)
            self.assertEqual(1, mock_save.call_count)
            self.assertIsInstance(result, pipeline.IndexItem)
            self.assertIsInstance(result.plugin,
                                  recordsets_plugin.RecordSetIndex)
            self.assertEqual(create_event[0], result.event_type)

    def test_delete_recordset(self):
        """Test recordset delete event"""
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        delete_event = (
            "dns.recordset.delete",
            {
                "zone_id": "3081593e-10ca-408c-af77-1397e689c177",
                "tenant_id": "80264096ac454d3d904002491fafe2ec",
                "created_at": "2016-03-31T05:48:53.000000",
                "description": None,
                "updated_at": "2016-03-31T07:24:14.000000",
                "records": [{
                    "status": "PENDING",
                    "zone_id": "3081593e-10ca-408c-af77-1397e689c177",
                    "managed": False,
                    "managed_resource_id": None,
                    "managed_resource_type": None,
                    "tenant_id": "80264096ac454d3d904002491fafe2ec",
                    "created_at": "2016-03-31T05:48:54.000000",
                    "managed_extra": None,
                    "updated_at": "2016-03-31T07:24:14.000000",
                    "managed_plugin_type": None,
                    "version": 3,
                    "managed_plugin_name": None,
                    "managed_tenant_id": None,
                    "action": "DELETE",
                    "hash": "8402992213c4c985429e9f2b23a18847",
                    "managed_resource_region": None,
                    "recordset_id": "cfe39618-49da-4877-914b-f25ef0fb3dc1",
                    "data": "10.0.0.1",
                    "id": "01835dc4-b805-4c8f-b962-2a96880f40c7",
                    "serial": 1459409054,
                    "description": None
                }],
                "version": 3,
                "ttl": None,
                "type": "A",
                "id": "cfe39618-49da-4877-914b-f25ef0fb3dc1",
                "name": "www.myzone.net."
            },
            "2016-03-31 07:24:14.945626"
        )

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        with mock.patch.object(self.plugin.index_helper,
                               'delete_document') as mock_delete:
            type_handler = event_handlers.get(delete_event[0], None)
            result = type_handler(*delete_event)
            self.assertEqual(1, mock_delete.call_count)
            self.assertIsInstance(result, pipeline.DeleteItem)
            self.assertIsInstance(result.plugin,
                                  recordsets_plugin.RecordSetIndex)
            self.assertEqual(delete_event[0], result.event_type)
            self.assertEqual(delete_event[1]["id"], result.doc_id)



# Copyright 2012 Bouvet ASA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import abc

from oslo_config import cfg
from oslo_log import log as logging
import six
from stevedore import extension


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


@six.add_metaclass(abc.ABCMeta)
class Plugin(object):
    """This class exists as a point for plugins to define
    config options.
    """
    def __init__(self):
        self.name = self.get_config_group_name()
        LOG.debug("Loaded plugin %s" % self.name)

    @classmethod
    def get_cfg_opts(cls):
        group = cfg.OptGroup(cls.get_config_group_name())
        opts = cls.get_plugin_opts()
        return [(group, opts)]

    @classmethod
    def get_config_group_name(cls):
        raise NotImplementedError()

    @classmethod
    def register_cfg_opts(cls, namespace):
        mgr = extension.ExtensionManager(namespace)

        for e in mgr:
            for group, opts in e.plugin.get_cfg_opts():
                if isinstance(group, six.string_types):
                    group = cfg.OptGroup(name=group)

                CONF.register_group(group)
                CONF.register_opts(opts, group=group)



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from searchlight.elasticsearch.plugins import utils as es_utils
from searchlight.tests import fake_plugins
from searchlight.tests import functional


class TestReindexing(functional.FunctionalTest):
    def __init__(self, *args, **kwargs):
        super(TestReindexing, self).__init__(*args, **kwargs)

    def setUp(self):
        super(TestReindexing, self).setUp()
        role_plugin = fake_plugins.RoleSeparatedPlugin(
            es_engine=self.elastic_connection)
        self.role_plugin = role_plugin

        non_role_plugin = fake_plugins.NonRoleSeparatedPlugin(
            es_engine=self.elastic_connection)
        self.non_role_plugin = non_role_plugin

        self.initialized_plugins['re-non-role-separated'] = non_role_plugin
        self.initialized_plugins['re-role-separated'] = role_plugin

    def create_es_documents(self, index):
        """Create a set of documents in ElasticSearch.
           Call all plugins used by the tests to dump their documents
           into ElasticSearch. This is the baseline for each test.
           There are 4 documents from the role plugin and 2 documents
           form the non-role plugin.
        """
        self.role_plugin.index_initial_data()
        self.non_role_plugin.index_initial_data()
        self.num_docs = 6
        self._flush_elasticsearch(index)

    def verify_initial_state(self):
        """Verify the initial state of ElasticSearch. This state is identical
           for all of the tests.
        """
        es_docs = self._get_all_elasticsearch_docs()
        self.assertEqual(self.num_docs, es_docs['hits']['total'])

    def verify_reindex_state(self, new_index):
        """Verify the state of ElasticSearch after the first reindex.
           This state is identical for all of the tests. At this point
           there should be a full set of documents both indices. The
           first check verifies both sets. The second check verifies
           the new index only.
        """
        es_docs = self._get_all_elasticsearch_docs()
        self.assertEqual((2 * self.num_docs), es_docs['hits']['total'])
        es_docs = self._get_all_elasticsearch_docs([new_index])
        self.assertEqual(self.num_docs, es_docs['hits']['total'])

    def verify_new_alias_state(self, new_index, alias_search, alias_listener):
        """Verify the state of ElasticSearch after the aliases are updated.
           This state is identical for all of the tests. At this point there
           should be only one set of documents. The first check verifies the
           number of documents. The next checks verify the aliases are correct.
        """
        es_docs = self._get_all_elasticsearch_docs([new_index])
        self.assertEqual(self.num_docs, es_docs['hits']['total'])
        es_alias = self._get_elasticsearch_aliases([new_index])
        self.assertEqual(1, len(es_alias))
        expected = {alias_listener: {}, alias_search: {}}
        self.assertEqual(es_alias[new_index]['aliases'], expected)

    def test_reindex_with_plugins(self):
        """Verify the reindexing functionality using both plugins to do the
        reindexing. We want to verify: the number of documents during reindex,
        the number of aliases during the reindex, the number of documents
        after the reindex and the aliases after the reindex.
        """
        alias_listener = self.role_plugin.alias_name_listener
        alias_search = self.role_plugin.alias_name_search
        resource_group = self.role_plugin.resource_group_name

        # Create a set of documents in ElasticSearch.
        self.create_es_documents(alias_listener)

        self.verify_initial_state()

        # Create and prepare a new index.
        new_index = es_utils.create_new_index(resource_group)
        self.role_plugin.prepare_index(index_name=new_index)
        self.non_role_plugin.prepare_index(index_name=new_index)

        # Set up the aliases.
        es_utils.setup_alias(new_index, alias_search, alias_listener)
        es_alias = self._get_elasticsearch_aliases([])
        self.assertEqual(2, len(es_alias))

        # Reindex using the plugins.
        self.role_plugin.index_initial_data()
        self.non_role_plugin.index_initial_data()
        self._flush_elasticsearch(alias_listener)

        self.verify_reindex_state(new_index)

        # Update aliases.
        old_index = es_utils.alias_search_update(alias_search, new_index)
        es_utils.delete_index(old_index)
        self._flush_elasticsearch(alias_listener)

        self.verify_new_alias_state(new_index=new_index,
                                    alias_search=alias_search,
                                    alias_listener=alias_listener)

    def test_reindex_with_es(self):
        """Verify the reindexing functionality using the elasticsearch reindex
        method to do the reindexing. We want to verify: the number of documents
        during reindex, the number of aliases during the reindex, the number of
        documents after the reindex and the aliases after the reindex.
        """
        alias_listener = self.role_plugin.alias_name_listener
        alias_search = self.role_plugin.alias_name_search
        resource_group = self.role_plugin.resource_group_name
        role_doc_type = self.role_plugin.document_type
        non_role_doc_type = self.non_role_plugin.document_type

        # Create a set of documents in ElasticSearch.
        self.create_es_documents(alias_listener)

        self.verify_initial_state()

        # Create and prepare a new index.
        new_index = es_utils.create_new_index(resource_group)
        self.role_plugin.prepare_index(index_name=new_index)
        self.non_role_plugin.prepare_index(index_name=new_index)

        # Set up the aliases.
        es_utils.setup_alias(new_index, alias_search, alias_listener)
        es_alias = self._get_elasticsearch_aliases([])
        self.assertEqual(2, len(es_alias))

        # Reindex using ElasticSearch.
        reindex = [role_doc_type, non_role_doc_type]
        es_utils.reindex(src_index=alias_listener, dst_index=new_index,
                         type_list=reindex)
        self._flush_elasticsearch(alias_listener)

        self.verify_reindex_state(new_index)

        # Update aliases.
        old_index = es_utils.alias_search_update(alias_search, new_index)
        es_utils.delete_index(old_index)
        self._flush_elasticsearch(alias_listener)

        self.verify_new_alias_state(new_index=new_index,
                                    alias_search=alias_search,
                                    alias_listener=alias_listener)

    def test_reindex_with_plugin_and_es(self):
        """Verify the reindexing functionality using both the plugin reindex
        and the elasticsearch reindex methods for the reindexing. We want to
        verify: the number of documents during reindex, the number of aliases
        during the reindex, the number of documents after the reindex and the
        aliases after the reindex.
        """
        alias_listener = self.role_plugin.alias_name_listener
        alias_search = self.role_plugin.alias_name_search
        resource_group = self.role_plugin.resource_group_name
        non_role_doc_type = self.non_role_plugin.document_type

        # Create a set of documents in ElasticSearch.
        self.create_es_documents(alias_listener)

        self.verify_initial_state()

        # Create and prepare a new index.
        new_index = es_utils.create_new_index(resource_group)
        self.role_plugin.prepare_index(index_name=new_index)
        self.non_role_plugin.prepare_index(index_name=new_index)

        # Set up the aliases.
        es_utils.setup_alias(new_index, alias_search, alias_listener)
        es_alias = self._get_elasticsearch_aliases([])
        self.assertEqual(2, len(es_alias))

        # Reindex. For role, use the plugin. For non-role use ElasticSearch.
        self.role_plugin.index_initial_data()
        reindex = [non_role_doc_type]
        es_utils.reindex(src_index=alias_listener, dst_index=new_index,
                         type_list=reindex)
        self._flush_elasticsearch(alias_listener)

        self.verify_reindex_state(new_index)

        # Update aliases.
        old_index = es_utils.alias_search_update(alias_search, new_index)
        es_utils.delete_index(old_index)
        self._flush_elasticsearch(alias_listener)

        self.verify_new_alias_state(new_index=new_index,
                                    alias_search=alias_search,
                                    alias_listener=alias_listener)



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from elasticsearch import helpers
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.helper import USER_ID_SUFFIX
from searchlight.elasticsearch.plugins.neutron import add_rbac
from searchlight.elasticsearch.plugins.neutron import serialize_floatingip
from searchlight.elasticsearch.plugins.neutron import serialize_network
from searchlight.elasticsearch.plugins.neutron import serialize_port
from searchlight.elasticsearch.plugins.neutron import serialize_router
from searchlight.elasticsearch.plugins.neutron import serialize_security_group
from searchlight.elasticsearch.plugins.neutron import serialize_subnet
from searchlight.elasticsearch.plugins import openstack_clients
from searchlight.elasticsearch.plugins import utils
from searchlight import pipeline

LOG = logging.getLogger(__name__)

SECGROUP_RETRIES = 20
RBAC_VALID_ACTIONS = ["access_as_shared", "access_as_external"]


class NetworkHandler(base.NotificationBase):
    @classmethod
    def _get_notification_exchanges(cls):
        return ['neutron']

    def get_event_handlers(self):
        return {
            'network.create.end': self.create_or_update,
            'network.update.end': self.create_or_update,
            'network.delete.end': self.delete,
            'rbac_policy.create.end': self.rbac_create,
            'rbac_policy.delete.end': self.rbac_delete
        }

    def get_log_fields(self, event_type, payload):
        # Delete and create/update notifications are different structures
        if 'network' in payload:
            return ('id', payload['network'].get('id')),
        elif 'network_id' in payload:
            return ('id', payload['network_id']),
        elif 'rbac_policy' in payload:
            return (
                ('network_id', payload['rbac_policy'].get('object_id')),
                ('target_tenant',
                 payload['rbac_policy'].get('target_tenant')),
                ('object_type',
                 payload['rbac_policy'].get('object_type')))
        elif 'rbac_policy_id' in payload:
            return ('rbac_policy_id', payload['rbac_policy_id']),
        return ()

    def create_or_update(self, event_type, payload, timestamp):
        network_id = payload['network']['id']
        LOG.debug("Updating network information for %s", network_id)

        # Neutron doesn't give us any date/time information
        network = serialize_network(payload['network'])
        version = self.get_version(network, timestamp)

        self.index_helper.save_document(network, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  network)

    def delete(self, event_type, payload, timestamp):
        network_id = payload['network_id']
        LOG.debug("Deleting network information for %s", network_id)
        try:
            # Note that it's not necessary to delete ports; neutron will not
            # allow deletion of a network that has ports assigned on it
            self.index_helper.delete_document({'_id': network_id})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       network_id)
        except Exception as exc:
            LOG.error(
                'Error deleting network %(network_id)s '
                'from index. Error: %(exc)s' %
                {'network_id': network_id, 'exc': exc})

    def rbac_create(self, event_type, payload, timestamp):
        """RBAC policy is making a network visible to users in a specific
           tenant. Previously this network was not visible to users in that
           tenant. We will want to add this tenant to the members list.
           Also add the RBAC policy.
        """
        valid_types = ["network"]

        event_type = payload['rbac_policy']['object_type']
        action = payload['rbac_policy']['action']
        if action not in RBAC_VALID_ACTIONS or event_type not in valid_types:
            # I'm bored. Nothing that concerns nor interests us.
            return

        network_id = payload['rbac_policy']['object_id']
        target_tenant = payload['rbac_policy']['target_tenant']
        policy_id = payload['rbac_policy']['id']
        LOG.debug("Adding RBAC policy for network %s with tenant %s",
                  network_id, target_tenant)

        # Read, modify, write an existing network document. Grab and modify
        # the admin version of the document. When saving the document it will
        # be indexed for both admin and user.
        doc = self.index_helper.get_document(network_id, for_admin=True)

        if not doc or not doc['_source']:
            LOG.error('Error adding rule to network. Network %(id)s '
                      'does not exist.' % {'id': network_id})
            return

        body = doc['_source']

        # Update network with RBAC policy.
        add_rbac(body, target_tenant, policy_id)

        # Bump version for race condition prevention. Use doc and not
        # body, since '_version' is outside of '_source'.
        version = doc['_version'] + 1
        self.index_helper.save_document(body, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  body)

    def rbac_delete(self, event_type, payload, timestamp):
        """RBAC policy is making a network invisible to users in specific
           tenant. Previously this network was visible to users in that
           tenant. We will remove this tenant from the members list.
           Also remove the RBAC policy.
        """
        policy_id = payload['rbac_policy_id']

        # Read, modify, write an existing network document. For both the
        # admin and user version of the document.

        # Find all documents (admin and user) with the policy ID.
        docs = self.index_helper.get_docs_by_nested_field(
            "rbac_policy", "rbac_id", policy_id, version=True)

        if not docs or not docs['hits']['hits']:
            return

        for doc in docs['hits']['hits']:
            if doc['_id'].endswith(USER_ID_SUFFIX):
                # We only want to use the admin document.
                continue
            body = doc['_source']

            target_tenant = None
            policies = body['rbac_policy']
            for p in policies:
                if p.get('rbac_id') == policy_id:
                    target_tenant = p['target_tenant']

            # Remove target_tenant from members list.
            members_list = (body['members'])
            if target_tenant in members_list:
                members_list.remove(target_tenant)
                body['members'] = members_list

            # Remove RBAC policy.
            new_list = [p for p in policies if p.get('rbac_id') != policy_id]
            body['rbac_policy'] = new_list

            # Bump version for race condition prevention. Use doc and not
            # body, since '_version' is outside of '_source'.
            version = doc['_version'] + 1
            self.index_helper.save_document(body, version=version)
            return pipeline.IndexItem(self.index_helper.plugin,
                                      event_type,
                                      payload,
                                      body)


class PortHandler(base.NotificationBase):
    @classmethod
    def _get_notification_exchanges(cls):
        return ['neutron']

    def get_event_handlers(self):
        return {
            'port.create.end': self.create_or_update,
            'port.update.end': self.create_or_update,
            'port.delete.end': self.delete,

            'router.interface.create': self.create_or_update_from_interface,
            'router.interface.delete': self.delete_from_interface
        }

    def get_log_fields(self, event_type, payload):
        # Delete and create/update notifications are different structures, and
        # network_id will not be present for delete notifications
        if event_type.startswith('router.interface'):
            port_id = payload.get('router_interface', {}).get('port_id')
            network_id = payload.get('router_interface', {}).get('network_id')
        else:
            port_id = payload.get('port_id',
                                  payload.get('port', {}).get('id'))
            network_id = payload.get('port', {}).get('network_id')

        return (
            ('id', port_id),
            ('network_id', network_id)
        )

    def create_or_update(self, event_type, payload, timestamp):
        port_id = payload['port']['id']

        if payload['port'].get('device_owner', None) == 'network:dhcp':
            # TODO(sjmc7): Remove this once we can get proper notifications
            # about DHCP ports.
            #  See https://bugs.launchpad.net/searchlight/+bug/1558790
            LOG.info("Skipping notification for DHCP port %s. If neutron "
                     "is sending notifications for DHCP ports, the "
                     "Searchlight plugin should be updated to process "
                     "them.", port_id)
            return

        LOG.debug("Updating port information for %s", port_id)

        # Version doesn't really make a huge amount of sense here but
        # is better than nothing
        port = serialize_port(payload['port'])
        version = self.get_version(port, timestamp)

        self.index_helper.save_document(port, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  port)

    def delete_port(self, event_type, payload, timestamp):
        return self.delete(event_type, payload, payload['port_id'])

    def delete(self, event_type, payload, port_id):
        LOG.debug("Deleting port information for %s; finding routing", port_id)
        try:
            self.index_helper.delete_document_unknown_parent(port_id)
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       port_id)
        except Exception as exc:
            LOG.error(
                'Error deleting port %(port_id)s '
                'from index. Error: %(exc)s' %
                {'port_id': port_id, 'exc': exc})

    def create_or_update_from_interface(self, event_type, payload, timestamp):
        """Unfortunately there seems to be no notification for ports created
        as part of a router interface creation, nor for DHCP ports. This
        means we need to go to the API.
        """
        port_id = payload['router_interface']['port_id']
        LOG.debug("Retrieving port %s from API", port_id)
        nc = openstack_clients.get_neutronclient()
        port = nc.show_port(port_id)['port']
        serialized = serialize_port(port)
        version = self.get_version(serialized, timestamp)
        self.index_helper.save_document(serialized, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  serialized
                                  )

    def delete_from_interface(self, event_type, payload, timestamp):
        """The partner of create_or_update_from_interface. There's no separate
        port deletion notification.
        """
        port_id = payload['router_interface']['port_id']
        LOG.debug("Deleting port %s from router interface", port_id)
        return self.delete(event_type, payload, port_id)


class SubnetHandler(base.NotificationBase):
    @classmethod
    def _get_notification_exchanges(cls):
        return ['neutron']

    def get_event_handlers(self):
        return {
            'subnet.create.end': self.create_or_update,
            'subnet.update.end': self.create_or_update,
            'subnet.delete.end': self.delete
        }

    def get_log_fields(self, event_type, payload):
        # Delete and create/update notifications are different structures, and
        # network_id will not be present for delete notifications
        subnet_id = payload.get('subnet_id',
                                payload.get('subnet', {}).get('id'))
        network_id = payload.get('subnet', {}).get('network_id')

        return (
            ('id', subnet_id),
            ('network_id', network_id)
        )

    def create_or_update(self, event_type, payload, timestamp):
        subnet_id = payload['subnet']['id']
        LOG.debug("Updating subnet information for %s", subnet_id)
        subnet = serialize_subnet(payload['subnet'])

        version = self.get_version(subnet, timestamp)
        self.index_helper.save_document(subnet, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  subnet)

    def delete(self, event_type, payload, timestamp):
        subnet_id = payload['subnet_id']
        LOG.debug("Deleting subnet information for %s; finding routing",
                  subnet_id)
        try:
            self.index_helper.delete_document_unknown_parent(subnet_id)
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       subnet_id)
        except Exception as exc:
            LOG.error(
                'Error deleting subnet %(subnet_id)s '
                'from index: %(exc)s' %
                {'subnet_id': subnet_id, 'exc': exc})


class RouterHandler(base.NotificationBase):
    @classmethod
    def _get_notification_exchanges(cls):
        return ['neutron']

    def get_event_handlers(self):
        return {
            'router.create.end': self.create_or_update,
            'router.update.end': self.create_or_update,
            'router.delete.end': self.delete
        }

    def get_log_fields(self, event_type, payload):
        # Delete and create/update notifications are different structures
        router_id = payload.get('router_id',
                                payload.get('router', {}).get('id'))
        return ('id', router_id),

    def create_or_update(self, event_type, payload, timestamp):
        router_id = payload['router']['id']
        LOG.debug("Updating router information for %s", router_id)
        router = serialize_router(
            payload['router'],
            updated_at=utils.timestamp_to_isotime(timestamp))
        version = self.get_version(router, timestamp)
        self.index_helper.save_document(router, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  router)

    def delete(self, event_type, payload, timestamp):
        router_id = payload['router_id']
        LOG.debug("Deleting router information for %s", router_id)
        try:
            self.index_helper.delete_document({'_id': router_id})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       router_id)
        except Exception as exc:
            LOG.error(
                'Error deleting router %(router)s '
                'from index: %(exc)s' %
                {'router': router_id, 'exc': exc})


class FloatingIPHandler(base.NotificationBase):
    @classmethod
    def _get_notification_exchanges(cls):
        return ['neutron']

    def get_event_handlers(self):
        return {
            'floatingip.create.end': self.create_or_update,
            'floatingip.update.end': self.create_or_update,
            'floatingip.delete.end': self.delete
        }

    def get_log_fields(self, event_type, payload):
        # Delete and create/update notifications are different structures
        fip_id = payload.get('floatingip_id',
                             payload.get('floatingip', {}).get('id'))
        network_id = payload.get('floatingip', {}).get('floating_network_id')
        return (
            ('id', fip_id),
            ('network_id', network_id)
        )

    def create_or_update(self, event_type, payload, timestamp):
        fip_id = payload['floatingip']['id']
        LOG.debug("Updating floatingip information for %s", fip_id)
        floatingip = serialize_floatingip(
            payload['floatingip'],
            updated_at=utils.timestamp_to_isotime(timestamp))
        version = self.get_version(floatingip, timestamp)
        self.index_helper.save_document(floatingip, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  floatingip)

    def delete(self, event_type, payload, timestamp):
        fip_id = payload['floatingip_id']
        LOG.debug("Deleting floatingip information for %s", fip_id)
        try:
            self.index_helper.delete_document({'_id': fip_id})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       fip_id)
        except Exception as exc:
            LOG.error(
                'Error deleting floating ip %(fip)s '
                'from index: %(exc)s' %
                {'fip': fip_id, 'exc': exc})


class SecurityGroupHandler(base.NotificationBase):
    @classmethod
    def _get_notification_exchanges(cls):
        return ['neutron']

    def get_event_handlers(self):
        return {
            'security_group.create.end': self.create_or_update_group,
            'security_group.delete.end': self.delete_group,
            'security_group_rule.create.end': self.create_or_update_rule,
            'security_group_rule.delete.end': self.delete_rule,
        }

    def get_log_fields(self, event_type, payload):
        if event_type == 'security_group_rule.create.end':
            sgr_payload = payload['security_group_rule']
            return (('id', sgr_payload['security_group_id']),
                    ('security_group_rule_id', sgr_payload['id']))
        elif event_type == 'security_group_rule.delete.end':
            return ('id', payload['security_group_rule_id']),

        group_id = payload.get('security_group_id',
                               payload.get('security_group', {}).get('id'))
        return ('id', group_id),

    def create_or_update_group(self, event_type, payload, timestamp):
        group_name = payload['security_group']['name']
        sec_id = payload['security_group']['id']
        LOG.debug("Updating security group information for %(grp)s (%(sec)s)" %
                  {'grp': group_name, 'sec': sec_id})

        # Version doesn't really make sense for security groups,
        # but we need to normalize the fields.
        doc = serialize_security_group(payload['security_group'])
        version = self.get_version(doc, timestamp)

        self.index_helper.save_document(doc, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  doc)

    def delete_group(self, event_type, payload, timestamp):
        sec_id = payload['security_group_id']
        LOG.debug("Deleting security group information for %s", sec_id)
        try:
            self.index_helper.delete_document({'_id': sec_id})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       sec_id)
        except Exception as exc:
            LOG.error(
                'Error deleting security_group %(sec_id)s. Error: %(exc)s' %
                {'sec_id': sec_id, 'exc': exc})

    def create_or_update_rule(self, event_type, payload, timestamp):
        # The issue here is that the notification is not complete.
        # We have only a single rule that needs to be added to an
        # existing group. A major issue is that we may be updating
        # the ES document while other workers are modifying the rules
        # in the same ES document. This requires an aggressive retry policy,
        # using the "version" field. Since the ES document will have been
        # modified after a conflict, we will need to grab the latest version
        # of the document before continuing. After "retries" number of times,
        # we will admit failure and not try the update anymore.
        # NB: Most of the looping logic is the same as in "delete_rule".
        #     The read/modify the ES document is different. If the logic
        #     changes, please make the changes there.
        group_id = payload['security_group_rule']['security_group_id']
        LOG.debug("Updating security group rule information for %s", group_id)

        for attempts in range(SECGROUP_RETRIES):
            # Read, modify, write of an existing security group.
            doc = self.index_helper.get_document(group_id)

            if not doc:
                return
            body = doc['_source']
            if not body or 'security_group_rules' not in body:
                return

            body['security_group_rules'].append(payload['security_group_rule'])

            version = doc['_version']
            try:
                version += 1
                self.index_helper.save_document(body, version=version)
                return pipeline.IndexItem(self.index_helper.plugin,
                                          event_type,
                                          payload,
                                          body)
            except helpers.BulkIndexError as e:
                if e.errors[0]['index']['status'] == 409:
                    # Conflict error, retry with new version of doc.
                    pass
                else:
                    raise

        if attempts == (SECGROUP_RETRIES - 1):
            LOG.error('Error adding security group rule %(id)s:'
                      ' Too many retries' % {'id': group_id})

    def delete_rule(self, event_type, payload, timestamp):
        # See comment for create_or_update_rule() for details.
        rule_id = payload['security_group_rule_id']
        LOG.debug("Updating security group rule information for %s", rule_id)

        field = 'security_group_rules'

        # Read, modify, write of an existing security group.
        # To avoid a race condition, we are searching for the document
        # in a round-about way. Outside of the retry loop, we will
        # search for the document and save the document ID. This way we
        # do not need to search inside the loop. We will access the document
        # directly by the ID which will always return the latest version.
        orig_doc = self.index_helper.get_docs_by_nested_field(
            "security_group_rules", "id", rule_id, version=True)
        if not orig_doc:
            return
        doc_id = orig_doc['hits']['hits'][0]['_id']
        doc = orig_doc['hits']['hits'][0]
        for attempts in range(SECGROUP_RETRIES):
            body = doc['_source']
            if not body or field not in body:
                return

            body[field] = \
                list(filter(lambda r: r['id'] != rule_id, body[field]))

            version = doc['_version']
            try:
                version += 1
                self.index_helper.save_document(body, version=version)
                return pipeline.IndexItem(self.index_helper.plugin,
                                          event_type,
                                          payload,
                                          body)
            except helpers.BulkIndexError as e:
                if e.errors[0]['index']['status'] == 409:
                    # Conflict. Retry with new version.
                    doc = self.index_helper.get_document(doc_id)
                    if not doc:
                        return
                else:
                    raise

        if attempts == (SECGROUP_RETRIES - 1):
            LOG.error('Error deleting security group rule %(id)s:'
                      ' Too many retries' % {'id': rule_id})



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base test class for running non-stubbed tests (functional tests)

The FunctionalTest class contains helper methods for starting the API
server, grabbing the logs of each, cleaning up pidfiles,
and spinning down the servers.
"""

import atexit
import datetime
import elasticsearch
import httplib2
import importlib
import json
import logging
import mock
import os
import platform
import shutil
import signal
import six
import socket
import sys
import tempfile
import time

import fixtures
from oslo_serialization import jsonutils
# NOTE(jokke): simplified transition to py3, behaves like py2 xrange
from six.moves import range
from six.moves import urllib
import testtools

from searchlight.common import utils
from searchlight.elasticsearch.plugins import utils as es_utils
from searchlight.tests import utils as test_utils


execute, get_unused_port = test_utils.execute, test_utils.get_unused_port
tracecmd_osmap = {'Linux': 'strace', 'FreeBSD': 'truss'}


class Server(object):
    """
    Class used to easily manage starting and stopping
    a server during functional test runs.
    """
    def __init__(self, test_dir, port, sock=None):
        """
        Creates a new Server object.

        :param test_dir: The directory where all test stuff is kept. This is
                         passed from the FunctionalTestCase.
        :param port: The port to start a server up on.
        """
        self.debug = True
        self.no_venv = False
        self.test_dir = test_dir
        self.bind_port = port
        self.conf_file_name = None
        self.conf_base = None
        self.paste_conf_base = None
        self.exec_env = None
        self.deployment_flavor = ''
        self.property_protection_file = ''
        self.log_file = None
        self.sock = sock
        self.fork_socket = True
        self.process_pid = None
        self.server_module = None
        self.stop_kill = False
        self.elasticsearch_pid = None

    def write_conf(self, **kwargs):
        """
        Writes the configuration file for the server to its intended
        destination.  Returns the name of the configuration file and
        the over-ridden config content (may be useful for populating
        error messages).
        """
        if not self.conf_base:
            raise RuntimeError("Subclass did not populate config_base!")

        conf_override = self.__dict__.copy()
        if kwargs:
            conf_override.update(**kwargs)

        # A config file and paste.ini to use just for this test...we don't want
        # to trample on currently-running Searchlight servers, now do we?

        conf_dir = os.path.join(self.test_dir, 'etc')
        conf_filepath = os.path.join(conf_dir, "%s.conf" % self.server_name)
        if os.path.exists(conf_filepath):
            os.unlink(conf_filepath)
        paste_conf_filepath = conf_filepath.replace(".conf", "-paste.ini")
        if os.path.exists(paste_conf_filepath):
            os.unlink(paste_conf_filepath)
        utils.safe_mkdirs(conf_dir)

        def override_conf(filepath, overridden):
            with open(filepath, 'w') as conf_file:
                conf_file.write(overridden)
                conf_file.flush()
                return conf_file.name

        overridden_core = self.conf_base % conf_override
        self.conf_file_name = override_conf(conf_filepath, overridden_core)

        overridden_paste = ''
        if self.paste_conf_base:
            overridden_paste = self.paste_conf_base % conf_override
            override_conf(paste_conf_filepath, overridden_paste)

        overridden = ('==Core config==\n%s\n==Paste config==\n%s' %
                      (overridden_core, overridden_paste))

        return self.conf_file_name, overridden

    def start(self, expect_exit=True, expected_exitcode=0, **kwargs):
        """
        Starts the server.

        Any kwargs passed to this method will override the configuration
        value in the conf file used in starting the servers.
        """

        # Ensure the configuration file is written
        self.write_conf(**kwargs)

        elasticsearch_wrapper.ensure_elasticsearch()

        cmd = ("%(server_module)s --config-file %(conf_file_name)s"
               % {"server_module": self.server_module,
                  "conf_file_name": self.conf_file_name})
        cmd = "%s -m %s" % (sys.executable, cmd)
        # close the sock and release the unused port closer to start time
        if self.exec_env:
            exec_env = self.exec_env.copy()
        else:
            exec_env = {}
        pass_fds = set()
        if self.sock:
            if not self.fork_socket:
                self.sock.close()
                self.sock = None
            else:
                fd = os.dup(self.sock.fileno())
                exec_env[utils.SEARCHLIGHT_TEST_SOCKET_FD_STR] = str(fd)
                pass_fds.add(fd)
                self.sock.close()

        self.process_pid = test_utils.fork_exec(cmd,
                                                logfile=os.devnull,
                                                exec_env=exec_env,
                                                pass_fds=pass_fds)

        self.stop_kill = not expect_exit
        if self.pid_file:
            pf = open(self.pid_file, 'w')
            pf.write('%d\n' % self.process_pid)
            pf.close()
        if not expect_exit:
            rc = 0
            try:
                os.kill(self.process_pid, 0)
            except OSError:
                raise RuntimeError("The process did not start")
        else:
            rc = test_utils.wait_for_fork(
                self.process_pid,
                expected_exitcode=expected_exitcode)
        # avoid an FD leak
        if self.sock:
            os.close(fd)
            self.sock = None
        return (rc, '', '')

    def reload(self, expect_exit=True, expected_exitcode=0, **kwargs):
        """
        Start and stop the service to reload

        Any kwargs passed to this method will override the configuration
        value in the conf file used in starting the servers.
        """
        self.stop()
        return self.start(expect_exit=expect_exit,
                          expected_exitcode=expected_exitcode, **kwargs)

    def stop(self):
        """
        Spin down the server.
        """
        if not self.process_pid:
            raise Exception('why is this being called? %s' % self.server_name)

        if self.stop_kill:
            os.kill(self.process_pid, signal.SIGTERM)
        rc = test_utils.wait_for_fork(self.process_pid, raise_error=False)
        return (rc, '', '')

    def dump_log(self, name):
        log = logging.getLogger(name)
        if not self.log_file or not os.path.exists(self.log_file):
            return
        fptr = open(self.log_file, 'r')
        for line in fptr:
            log.info(line.strip())


class SearchServer(Server):

    """
    Server object that starts/stops/manages the Search server.
    To be decided: where elasticsearch comes from.
    """

    def __init__(self, test_dir, port, policy_file, sock=None):
        super(SearchServer, self).__init__(test_dir, port, sock=sock)
        self.server_name = 'api'
        self.server_module = 'searchlight.cmd.%s' % self.server_name

        self.elasticsearch_port = elasticsearch_wrapper.elasticsearch_port
        self.pid_file = os.path.join(self.test_dir, "searchlight.pid")
        self.log_file = os.path.join(self.test_dir, "searchlight.log")
        self.api_version = 1
        self.policy_file = policy_file
        self.policy_default_rule = 'default'
        self.property_protection_rule_format = 'roles'

        self.service_policy_files = ''
        self.service_policy_path = ''

        self.conf_base = """[DEFAULT]
debug = %(debug)s
log_file = %(log_file)s
api_limit_max = 1000
policy_file = %(policy_file)s
policy_default_rule = %(policy_default_rule)s

property_protection_file = %(property_protection_file)s
property_protection_rule_format = %(property_protection_rule_format)s

[paste_deploy]
flavor = %(deployment_flavor)s

[elasticsearch]
hosts = 127.0.0.1:%(elasticsearch_port)s

[service_policies]
service_policy_files = %(service_policy_files)s
service_policy_path = %(service_policy_path)s

[api]
# Hardcoding a single worker; the cleanup code can't deal with more
workers = 0
bind_host = 127.0.0.1
bind_port = %(bind_port)s

"""
        self.paste_conf_base = """[pipeline:searchlight]
pipeline = versionnegotiation unauthenticated-context rootapp

[pipeline:searchlight-keystone]
pipeline = versionnegotiation authtoken context rootapp

[pipeline:searchlight-trusted-auth]
pipeline = context rootapp

[composite:rootapp]
paste.composite_factory = searchlight.api:root_app_factory
/: apiversions
/v1: apiv1app

[app:apiversions]
paste.app_factory = searchlight.api.versions:create_resource

[app:apiv1app]
paste.app_factory = searchlight.api.v1.router:API.factory

[filter:versionnegotiation]
paste.filter_factory =
 searchlight.api.middleware.version_negotiation:VersionNegotiationFilter.factory

[filter:unauthenticated-context]
paste.filter_factory =
 searchlight.api.middleware.context:UnauthenticatedContextMiddleware.factory

[filter:authtoken]
paste.filter_factory = keystonemiddleware.auth_token:filter_factory
delay_auth_decision = true

[filter:context]
paste.filter_factory =
 searchlight.api.middleware.context:ContextMiddleware.factory
"""


class FunctionalTest(test_utils.BaseTestCase):

    """
    Base test class for any test that wants to test the actual
    servers and clients and not just the stubbed out interfaces
    """

    inited = False
    disabled = False
    launched_servers = []

    @test_utils.depends_on_exe("elasticsearch")
    @test_utils.skip_if_disabled
    def setUp(self):
        super(FunctionalTest, self).setUp()
        self.test_dir = self.useFixture(fixtures.TempDir()).path

        self.api_protocol = 'http'
        self.api_port, search_sock = test_utils.get_unused_port_and_socket()

        self.tracecmd = tracecmd_osmap.get(platform.system())

        self.conf_dir = os.path.join(self.test_dir, 'etc')
        utils.safe_mkdirs(self.conf_dir)
        self.copy_data_file('policy.json', self.conf_dir)
        self.copy_data_file('property-protections.conf', self.conf_dir)
        self.copy_data_file('property-protections-policies.conf',
                            self.conf_dir)
        self.property_file_roles = os.path.join(self.conf_dir,
                                                'property-protections.conf')
        property_policies = 'property-protections-policies.conf'
        self.property_file_policies = os.path.join(self.conf_dir,
                                                   property_policies)
        self.policy_file = os.path.join(self.conf_dir, 'policy.json')

        self.api_server = SearchServer(self.test_dir,
                                       self.api_port,
                                       self.policy_file,
                                       sock=search_sock)
        self._additional_server_config()

        self.pid_files = [self.api_server.pid_file]
        self.files_to_destroy = []
        self.launched_servers = []

        self.elastic_connection = elasticsearch.Elasticsearch(
            "http://localhost:%s" % self.api_server.elasticsearch_port)

        self.api_server.deployment_flavor = "trusted-auth"
        # Use the role-based policy file all over; we need it for the property
        # protection tests
        self.api_server.property_protection_file = self.property_file_roles

        self.base_url = "http://127.0.0.1:%d/v1" % self.api_port
        self.start_with_retry(self.api_server,
                              "api_port",
                              max_retries=3,
                              **self.__dict__.copy())
        self.addCleanup(self.cleanup)

        self.initialized_plugins = {}
        self.configurePlugins()

    def _additional_server_config(self):
        """Any additional configuration for the API server. This is run prior
        to writing the server config file and starting the server.
        """
        pass

    def configurePlugins(self, include_plugins=None, exclude_plugins=()):
        """Specify 'exclude_plugins' or 'include_plugins' as a list of
        tuples.
        """
        plugin_classes = {
            'glance': {'images': 'ImageIndex', 'metadefs': 'MetadefIndex'},
            'nova': {'servers': 'ServerIndex',
                     'hypervisors': 'HypervisorIndex',
                     'flavors': 'FlavorIndex',
                     'servergroups': 'ServerGroupIndex'},
            'cinder': {'volumes': 'VolumeIndex', 'snapshots': 'SnapshotIndex'},
            'neutron': {'networks': 'NetworkIndex', 'ports': 'PortIndex',
                        'subnets': 'SubnetIndex', 'routers': 'RouterIndex',
                        'floatingips': 'FloatingIPIndex',
                        'security_groups': 'SecurityGroupIndex'},
            'swift': {'accounts': 'AccountIndex',
                      'containers': 'ContainerIndex',
                      'objects': 'ObjectIndex'},
            'designate': {'zones': 'ZoneIndex',
                          'recordsets': 'RecordSetIndex'}
        }

        plugins = include_plugins or (
            ('glance', 'images'), ('glance', 'metadefs'),
            ('nova', 'servers'), ('nova', 'hypervisors'),
            ('nova', 'flavors'), ('nova', 'servergroups'),
            ('cinder', 'volumes'), ('cinder', 'snapshots'),
            ('neutron', 'networks'), ('neutron', 'ports'),
            ('neutron', 'subnets'), ('neutron', 'routers'),
            ('neutron', 'floatingips'), ('neutron', 'security_groups'),
            ('cinder', 'volumes'), ('cinder', 'snapshots'),
            ('swift', 'accounts'), ('swift', 'containers'),
            ('swift', 'objects'),
            ('designate', 'zones'), ('designate', 'recordsets')
        )
        plugins = filter(lambda plugin: plugin not in exclude_plugins, plugins)

        # Make sure the plugins instantiated in this process use the same
        # connection as the ones in the API process they'll work with
        es_conn_patcher = mock.patch('searchlight.elasticsearch.get_api',
                                     return_value=self.elastic_connection)
        es_conn_patcher.start()
        self.addCleanup(es_conn_patcher.stop)

        index_name = es_utils.create_new_index('searchlight')

        for service, plugin_type in plugins:
            plugin_mod_name = ("searchlight.elasticsearch.plugins.%s.%s"
                               % (service, plugin_type))
            plugin_cls_name = plugin_classes[service][plugin_type]

            plugin_mod = importlib.import_module(plugin_mod_name)
            plugin_cls = getattr(plugin_mod, plugin_cls_name)

            # This'll call our dummy init (above)
            plugin_instance = plugin_cls()

            self.initialized_plugins[plugin_instance.document_type] = \
                plugin_instance

        # Reproduce the logic from searchlight.common.utils to set up
        # parent/child relationships; the stevedore structure is different
        for instance in self.initialized_plugins.values():
            parent_plugin_name = instance.parent_plugin_type()
            if parent_plugin_name:
                parent_plugin = self.initialized_plugins[parent_plugin_name]
                instance.register_parent(parent_plugin)

        # Reproduce the logic from cmd.manage to prepare the index.
        for instance in self.initialized_plugins.values():
            instance.prepare_index(index_name=index_name)

        # Create the aliases
        es_utils.setup_alias(index_name, 'searchlight-search',
                             'searchlight-listener')

    def tearDown(self):
        super(FunctionalTest, self).tearDown()

        self.api_server.dump_log('api_server')

    def _index(self, plugin, docs, refresh_index=True):
        """Index data exactly as the plugin would under searchlight-manage.
        docs must be an iterable of whatever the plugin's 'get_objects' call
        would return.
        """
        with mock.patch.object(plugin, 'get_objects', return_value=docs):
            with mock.patch.object(plugin, 'child_plugins', return_value=[]):
                if hasattr(plugin, 'get_rbac_objects'):
                    rbac_mock = mock.patch.object(plugin, 'get_rbac_objects',
                                                  return_value={})
                    rbac_mock.start()
                plugin.index_initial_data()

        if refresh_index:
            # Force elasticsearch to update its search index
            self._flush_elasticsearch(plugin.alias_name_listener)

    def _flush_elasticsearch(self, index_name=None):
        self.elastic_connection.indices.flush(index_name)

    def _headers(self, tenant_id, custom_headers={}):
        base_headers = {
            "X-Identity-Status": "Confirmed",
            "X-Auth-Token": "932c5c84-02ac-4fe5-a9ba-620af0e2bb96",
            "X-User-Id": "f9a41d13-0c13-47e9-bee2-ce4e8bfe958e",
            "X-Tenant-Id": tenant_id,
            "X-Roles": "member",
            "Content-Type": "application/json",
            "Accept": "application/json"
        }
        base_headers.update(custom_headers)
        return base_headers

    def _request(self, method, uri, tenant, body=None,
                 role="member", decode_json=True, extra_headers={}):
        custom_headers = {
            "X-Tenant-Id": tenant,
            "X-Roles": role,
        }
        custom_headers.update(extra_headers)
        headers = self._headers(tenant, custom_headers)
        if not tenant:
            del headers['X-Tenant-Id']
        kwargs = {
            "headers": headers
        }
        if body:
            kwargs["body"] = jsonutils.dumps(body)

        http = httplib2.Http()
        response, content = http.request(
            self.base_url + uri,
            method,
            **kwargs
        )
        if decode_json:
            content = jsonutils.loads(content)
        return response, content

    def _search_request(self, body, tenant, role="member", decode_json=True,
                        is_admin_project=None):
        """Conduct a search against all elasticsearch indices unless specified
        in `body`. Returns the response and json-decoded content.
        """
        extra_headers = {}
        if is_admin_project is not None:
            extra_headers['X-Is-Admin-Project'] = str(is_admin_project)

        return self._request("POST", "/search", tenant, body,
                             role, decode_json, extra_headers)

    def _facet_request(self, tenant, doc_type=None, role="member",
                       decode_json=True, include_fields=None,
                       exclude_options=None):
        url = "/search/facets"
        params = {}
        if doc_type:
            params['type'] = doc_type
        if include_fields is not None:
            params['include_fields'] = "true" if include_fields else "false"
        if exclude_options is not None:
            params['exclude_options'] = "true" if exclude_options else "false"
        url += '?' + urllib.parse.urlencode(params)
        return self._request("GET", url, tenant, role=role,
                             decode_json=decode_json)

    def _plugin_list_request(self, tenant, role="member"):
        url = "/search/plugins"
        return self._request("GET", url, tenant, role=role)

    def _get_hit_source(self, es_response):
        """Parse the _source from the elasticsearch hits"""
        if isinstance(es_response, six.string_types):
            es_response = jsonutils.loads(es_response)
        return [h["_source"] for h in es_response["hits"]["hits"]]

    def _get_all_elasticsearch_docs(self, indices=[]):
        """Query ES and return all documents. The caller can specify a list
           of indices to query. If the list is empty, we will query all
            indices.
        """
        index = ','.join(indices)
        es_url = "http://localhost:%s/%s/_search" % (
            self.api_server.elasticsearch_port, index)
        response, content = httplib2.Http().request(es_url)
        self.assertEqual(200, response.status)
        return jsonutils.loads(content)

    def _get_elasticsearch_doc(self, index_name, doc_type, doc_id):
        es_url = "http://localhost:%s/%s/%s/%s" % (
            self.api_server.elasticsearch_port, index_name, doc_type, doc_id)

        response, content = httplib2.Http().request(es_url)
        json_content = jsonutils.loads(content)
        return json_content

    def _delete_elasticsearch_doc(self, index_name, doc_type, doc_id):
        es_url = "http://localhost:%s/%s/%s/%s" % (
            self.api_server.elasticsearch_port, index_name, doc_type, doc_id)

        response, content = httplib2.Http().request(es_url, "DELETE")
        json_content = jsonutils.loads(content)
        return json_content

    def _get_elasticsearch_aliases(self, indices):
        """Return all aliases associated with a specified index(es). The
           caller can specify a list of indices. If the list is empty, we
           will query all indices.
        """
        index = ','.join(indices)
        es_url = "http://localhost:%s/%s/_aliases" % (
            self.api_server.elasticsearch_port, index)
        response, content = httplib2.Http().request(es_url)
        return jsonutils.loads(content)

    def _load_fixture_data(self, name):
        base_dir = "searchlight/tests/functional/data"
        with open(os.path.join(base_dir, name), 'r') as f:
            return json.load(f)

    def set_policy_rules(self, rules):
        fap = open(self.policy_file, 'w')
        fap.write(jsonutils.dumps(rules))
        fap.close()

    def cleanup(self):
        """
        Makes sure anything we created or started up in the
        tests are destroyed or spun down
        """

        # NOTE(jbresnah) call stop on each of the servers instead of
        # checking the pid file.  stop() will wait until the child
        # server is dead.  This eliminates the possibility of a race
        # between a child process listening on a port actually dying
        # and a new process being started
        servers = [self.api_server]
        for s in servers:
            try:
                s.stop()
            except Exception:
                pass

        for f in self.files_to_destroy:
            if os.path.exists(f):
                os.unlink(f)

        for plugin_instance in self.initialized_plugins.values():
            self.elastic_connection.indices.delete(
                index=plugin_instance.alias_name_search,
                ignore=404)
            self.elastic_connection.indices.delete(
                index=plugin_instance.alias_name_listener,
                ignore=404)

    def start_server(self,
                     server,
                     expect_launch,
                     expect_exit=True,
                     expected_exitcode=0,
                     **kwargs):
        """
        Starts a server on an unused port.

        Any kwargs passed to this method will override the configuration
        value in the conf file used in starting the server.

        :param server: the server to launch
        :param expect_launch: true iff the server is expected to
                              successfully start
        :param expect_exit: true iff the launched process is expected
                            to exit in a timely fashion
        :param expected_exitcode: expected exitcode from the launcher
        """
        self.cleanup()

        # Start up the requested server
        exitcode, out, err = server.start(expect_exit=expect_exit,
                                          expected_exitcode=expected_exitcode,
                                          **kwargs)
        if expect_exit:
            self.assertEqual(expected_exitcode, exitcode,
                             "Failed to spin up the requested server. "
                             "Got: %s" % err)

        self.launched_servers.append(server)

        launch_msg = self.wait_for_servers([server], expect_launch)
        self.assertIsNone(launch_msg, launch_msg)

    def start_with_retry(self, server, port_name, max_retries,
                         expect_launch=True,
                         **kwargs):
        """
        Starts a server, with retries if the server launches but
        fails to start listening on the expected port.

        :param server: the server to launch
        :param port_name: the name of the port attribute
        :param max_retries: the maximum number of attempts
        :param expect_launch: true iff the server is expected to
                              successfully start
        :param expect_exit: true iff the launched process is expected
                            to exit in a timely fashion
        """
        launch_msg = None
        for i in range(max_retries):
            exitcode, out, err = server.start(expect_exit=not expect_launch,
                                              **kwargs)
            name = server.server_name
            self.assertEqual(0, exitcode,
                             "Failed to spin up the %s server. "
                             "Got: %s" % (name, err))
            launch_msg = self.wait_for_servers([server], expect_launch)
            if launch_msg:
                server.stop()
                server.bind_port = get_unused_port()
                setattr(self, port_name, server.bind_port)
            else:
                self.launched_servers.append(server)
                break
        self.assertIsNone(launch_msg, launch_msg)

    def ping_server(self, port):
        """
        Simple ping on the port. If responsive, return True, else
        return False.

        :note We use raw sockets, not ping here, since ping uses ICMP and
        has no concept of ports...
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect(("127.0.0.1", port))
            s.close()
            return True
        except socket.error:
            return False

    def wait_for_servers(self, servers, expect_launch=True, timeout=10):
        """
        Tight loop, waiting for the given server port(s) to be available.
        Returns when all are pingable. There is a timeout on waiting
        for the servers to come up.

        :param servers: Searchlight server ports to ping
        :param expect_launch: Optional, true iff the server(s) are
                              expected to successfully start
        :param timeout: Optional, defaults to 3 seconds
        :return: None if launch expectation is met, otherwise an
                 assertion message
        """
        now = datetime.datetime.now()
        timeout_time = now + datetime.timedelta(seconds=timeout)
        replied = []
        while (timeout_time > now):
            pinged = 0
            for server in servers:
                if self.ping_server(server.bind_port):
                    pinged += 1
                    if server not in replied:
                        replied.append(server)
            if pinged == len(servers):
                msg = 'Unexpected server launch status'
                return None if expect_launch else msg
            now = datetime.datetime.now()
            time.sleep(0.05)

        failed = list(set(servers) - set(replied))
        msg = 'Unexpected server launch status for: '
        for f in failed:
            msg += ('%s, ' % f.server_name)
            if os.path.exists(f.pid_file):
                pid = f.process_pid
                trace = f.pid_file.replace('.pid', '.trace')
                if self.tracecmd:
                    cmd = '%s -p %d -o %s' % (self.tracecmd, pid, trace)
                    execute(cmd, raise_error=False, expect_exit=False)
                    time.sleep(0.5)
                    if os.path.exists(trace):
                        msg += ('\n%s:\n%s\n' % (self.tracecmd,
                                                 open(trace).read()))

        self.add_log_details(failed)

        return msg if expect_launch else None

    def stop_server(self, server, name):
        """
        Called to stop a single server in a normal fashion using the
        searchlight-control stop method to gracefully shut the server down.

        :param server: the server to stop
        """
        # Spin down the requested server
        server.stop()

    def stop_servers(self):
        """
        Called to stop the started servers in a normal fashion. Note
        that cleanup() will stop the servers using a fairly draconian
        method of sending a SIGTERM signal to the servers. Here, we use
        the searchlight-control stop method to gracefully shut the server down.
        This method also asserts that the shutdown was clean, and so it
        is meant to be called during a normal test case sequence.
        """

        # Spin down the API server
        self.stop_server(self.api_server, 'API server')

    def copy_data_file(self, file_name, dst_dir):
        src_file_name = os.path.join('searchlight/tests/etc', file_name)
        shutil.copy(src_file_name, dst_dir)
        dst_file_name = os.path.join(dst_dir, file_name)
        return dst_file_name

    def add_log_details(self, servers=None):
        logs = [s.log_file for s in (servers or self.launched_servers)]
        for log in logs:
            if os.path.exists(log):
                testtools.content.attach_file(self, log)


class ElasticsearchWrapper(object):
    """Helper to make sure elasticsearch is running. Tests should clear up
    after themselves because they'll be using a shared instance to avoid
    a ~5 second startup time
    """
    def __init__(self):
        self.elasticsearch_pid = None
        self._elasticsearch_port = None

    @property
    def elasticsearch_port(self):
        if self._elasticsearch_port is None:
            self._elasticsearch_port = get_unused_port()
        return self._elasticsearch_port

    def ensure_elasticsearch(self):
        """Start elasticsearch if it's not already running"""
        if self.elasticsearch_pid:
            # Use the existing instance
            return

        # Create a random directory and get or assign a port
        elasticsearch_root_dir = tempfile.mkdtemp(prefix='elastic')
        es_options = {
            'node.local': 'true',
            'index.number_of_shards': 1,
            'index.number_of_replicas': 0,
            'network.host': '127.0.0.1',
            'http.port': self.elasticsearch_port,
            'discovery.zen.ping.multicast.enabled': 'false',
            'path.data': elasticsearch_root_dir,
            'path.conf': elasticsearch_root_dir,
            'path.logs': elasticsearch_root_dir,
            'action.auto_create_index': 'false',
            'script.engine.groovy.inline.update': 'on'
        }
        # Set JVM options
        exec_env = {
            'ES_HEAP_SIZE': '64m'
        }
        cmd = 'elasticsearch '
        cmd += ' '.join('--%s=%s' % kv for kv in es_options.items())
        # Fork and retain the PID
        self.elasticsearch_pid = test_utils.fork_exec(cmd,
                                                      logfile=os.devnull,
                                                      exec_env=exec_env)

        # Register a shutdown function
        def _stop_elasticsearch():
            if not self.elasticsearch_pid:
                raise Exception('why is this being called? elasticsearch')
            os.kill(self.elasticsearch_pid, signal.SIGTERM)
            rc = test_utils.wait_for_fork(self.elasticsearch_pid,
                                          raise_error=False)
            self.elasticsearch_pid = None
            # Delete the temporary directory
            shutil.rmtree(elasticsearch_root_dir)
            return (rc, '', '')

        atexit.register(_stop_elasticsearch)

        # Wait for elasticsearch to spin up; it takes a while to initialize
        http = httplib2.Http()
        es_url = 'http://localhost:%s' % self.elasticsearch_port
        time.sleep(10)
        for _ in range(6):
            try:
                response, content = http.request(es_url)
                if response.status == 200:
                    break
            except socket.error:
                # Expect 'connection refused' a couple of times
                pass
            time.sleep(10)
        else:
            raise Exception("Elasticsearch failed to start")


elasticsearch_wrapper = ElasticsearchWrapper()



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.neutron import notification_handlers
from searchlight.elasticsearch.plugins.neutron import serialize_port
from searchlight.elasticsearch.plugins import openstack_clients


class PortIndex(base.IndexBase):
    NotificationHandlerCls = notification_handlers.PortHandler

    ADMIN_ONLY_FIELDS = ["binding:*"]

    @classmethod
    def get_document_type(self):
        return resource_types.NEUTRON_PORT

    @classmethod
    def parent_plugin_type(cls):
        return 'OS::Neutron::Net'

    def get_parent_id_field(self):
        return 'network_id'

    def get_mapping(self):
        string_not_analyzed = {'type': 'string', 'index': 'not_analyzed'}
        ordered_string = {'type': 'string',
                          'fields': {'raw': string_not_analyzed}}

        # TODO(sjmc7): denormalize 'device' fields (instance_id, router_id)?
        return {
            'dynamic': False,
            'properties': {
                'admin_state_up': {'type': 'boolean'},
                'binding:host_id': string_not_analyzed,
                'binding:vif_details': {'type': 'object', 'properties': {}},
                'binding:vif_type': string_not_analyzed,
                'binding:profile': {'type': 'object', 'properties': {}},
                'binding:vnic_type': string_not_analyzed,
                'created_at': {'type': 'date'},
                # device_owner and device_id identifies e.g which router
                # or instance 'owns' a port
                'device_id': string_not_analyzed,
                'device_owner': string_not_analyzed,
                'dns_name': string_not_analyzed,
                'extra_dhcp_opts': {
                    'type': 'nested',
                    'properties': {
                        'opt_name': string_not_analyzed,
                        'opt_value': string_not_analyzed
                    }
                },
                'fixed_ips': {
                    'type': 'nested',
                    'properties': {
                        'subnet_id': string_not_analyzed,
                        'ip_address': string_not_analyzed
                    }
                },
                'id': string_not_analyzed,
                'mac_address': string_not_analyzed,
                'name': ordered_string,
                'network_id': string_not_analyzed,
                'port_security_enabled': {'type': 'boolean'},
                'project_id': string_not_analyzed,
                'tenant_id': string_not_analyzed,
                'security_groups': string_not_analyzed,
                'status': string_not_analyzed,
                'updated_at': {'type': 'date'}
            },
            "_meta": {
                "project_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "tenant_id": {
                    "resource_type": resource_types.KEYSTONE_PROJECT
                },
                "fixed_ips.subnet_id": {
                    "resource_type": resource_types.NEUTRON_SUBNET
                }
            }
        }

    @property
    def admin_only_fields(self):
        from_conf = super(PortIndex, self).admin_only_fields
        return from_conf + PortIndex.ADMIN_ONLY_FIELDS

    @property
    def facets_with_options(self):
        return ('binding:vif_type', 'device_owner', 'admin_state_up', 'status',
                'binding:vnic_type', 'port_security_enabled')

    @property
    def facets_excluded(self):
        return {'tenant_id': True, 'project_id': True}

    @property
    def resource_allowed_policy_target(self):
        # Neutron only supports policy for individual resources
        return None

    @property
    def service_type(self):
        return 'network'

    def _get_rbac_field_filters(self, request_context):
        if request_context.is_admin:
            '''lakshmiS: neutron allows users with admin role to
            see ports of network(s) from other tenants when those
            networks are either shared or external.'''
            has_parent = {
                'has_parent': {
                    'type': self.parent_plugin_type(),
                    'query': {
                        "bool": {
                            "should": [
                                {'term': {'shared': True}},
                                {'term': {'router:external': True}}
                            ]
                        }
                    }
                }
            }

            return [
                {'term': {'tenant_id': request_context.owner}},
                has_parent  # from above
            ]
        else:
            ''' Users without admin role are not allowed to see the
            port information for networks from other tenants even if
            that network was shared or external.'''
            return [
                {'term': {'tenant_id': request_context.owner}}
            ]

    def get_objects(self):
        # This list of owners shamelessly taken from the Neutron code:
        #     neutron/neutron/notifiers/nova.py
        valid_owners = ('compute:', 'baremetal:')

        neutron_client = openstack_clients.get_neutronclient()
        for port in neutron_client.list_ports()['ports']:
            if (not port['device_owner'] or
               not port['device_owner'].startswith(valid_owners)):
                continue
            """The check/TODO below is superseded by the above check. When we
               flesh out the device owner check, re-enable the check below.
            # TODO(sjmc7): Remove this once we can get proper notifications
            # about DHCP ports.
            #  See https://bugs.launchpad.net/searchlight/+bug/1558790
            if port['device_owner'] == 'network:dhcp':
                continue
            """

            yield port

    def serialize(self, port):
        return serialize_port(port)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from distutils.version import LooseVersion
import os

from cinderclient import client as cinder_client
from designateclient.v2 import client as designate_client
from glanceclient import client as glance_client
from ironicclient import client as ironic_client
from ironicclient import exc as ironic_exceptions
from keystoneauth1 import loading as ka_loading
from keystoneclient import exceptions as keystone_exceptions

import keystoneclient.v3.client as ks_client
import neutronclient.v2_0.client as neutron_client
from novaclient import client as nova_client
import swiftclient

from searchlight.common import exception

from oslo_config import cfg


client_opts = [
    cfg.StrOpt('os-region-name',
               default=os.environ.get('OS_REGION_NAME'),
               help='Region name to use for OpenStack service endpoints.'
                    'If set, will be included in plugin mappings.'),
    cfg.StrOpt('os-endpoint-type',
               default=os.environ.get('OS_ENDPOINT_TYPE', 'publicURL'),
               help='Type of endpoint in Identity service catalog to '
                    'use for communication with OpenStack services.'),
]

compute_api_version = cfg.StrOpt(
    'compute_api_version',
    default='2.1',
    help='The compute API (micro)version, the provided '
         'compute API (micro)version should be not smaller '
         'than 2.1 and not larger than the max supported '
         'Compute API microversion. The current supported '
         'Compute API versions can be checked using: '
         'nova version-list.')

GROUP = "service_credentials"
CONF = cfg.CONF

CONF.register_opts(client_opts, group=GROUP)
CONF.register_opt(compute_api_version, group="service_credentials:nova")

ka_loading.register_session_conf_options(CONF, GROUP)
ka_loading.register_auth_conf_options(CONF, GROUP)

_session = None

NOVA_MIN_API_VERSION = '2.1'
IRONIC_API_VERSION = '1.22'


def _get_session():
    global _session
    if not _session:
        auth = ka_loading.load_auth_from_conf_options(CONF, GROUP)
        _session = ka_loading.load_session_from_conf_options(
            CONF, GROUP, auth=auth)
    return _session


def get_glanceclient():
    session = _get_session()

    return glance_client.Client(
        version='2',
        session=session,
        interface=CONF.service_credentials.os_endpoint_type,
        region_name=CONF.service_credentials.os_region_name
    )


def get_novaclient():

    def do_get_client(api_version=2.1):
        session = _get_session()
        return nova_client.Client(
            version=api_version,
            session=session,
            region_name=CONF.service_credentials.os_region_name,
            endpoint_type=CONF.service_credentials.os_endpoint_type
        )

    version = CONF["service_credentials:nova"].compute_api_version
    # Check whether Nova can support the provided microversion.
    max_version = do_get_client().versions.list()[-1].version
    if LooseVersion(version) > LooseVersion(max_version) or \
            LooseVersion(version) < LooseVersion(NOVA_MIN_API_VERSION):
        raise exception.InvalidAPIVersionProvided(
            service='compute service', min_version=NOVA_MIN_API_VERSION,
            max_version=max_version)

    return do_get_client(version)


def get_designateclient():
    session = _get_session()

    return designate_client.Client(
        session=session,
        region_name=CONF.service_credentials.os_region_name,
        endpoint_type=CONF.service_credentials.os_endpoint_type
    )


def get_neutronclient():
    session = _get_session()
    return neutron_client.Client(
        session=session,
        region_name=CONF.service_credentials.os_region_name,
        endpoint_type=CONF.service_credentials.os_endpoint_type
    )


def get_cinderclient():
    session = _get_session()

    return cinder_client.Client(
        version='2',
        session=session,
        region_name=CONF.service_credentials.os_region_name,
        endpoint_type=CONF.service_credentials.os_endpoint_type
    )

# Swift still needs special handling because it doesn't support
# keystone sessions. Rather than maintain two codepaths, we'll do this
_swiftclient = None


def clear_cached_swiftclient_on_unauthorized(fn):
    def wrapper(*args, **kwargs):
        global _session
        global _swiftclient
        try:
            return fn(*args, **kwargs)
        except swiftclient.exceptions.ClientException:
            _session = None
            _swiftclient = None
            return fn(*args, **kwargs)
    return wrapper


def get_swiftclient():

    global _swiftclient
    if _swiftclient:
        return _swiftclient
    _get_session()

    service_type = 'object-store'

    os_options = {
        'service_type': service_type,
        'region_name': CONF.service_credentials.os_region_name,
        'endpoint_type': CONF.service_credentials.os_endpoint_type,
    }

    # When swiftclient supports session, use session instead of
    # preauthtoken param below
    _swiftclient = swiftclient.client.Connection(
        auth_version='2',
        user=CONF.service_credentials.username,
        key=CONF.service_credentials.password,
        authurl=CONF.service_credentials.auth_url,
        tenant_name=CONF.service_credentials.tenant_name,
        os_options=os_options,
        cacert=CONF.service_credentials.cafile,
        insecure=CONF.service_credentials.insecure
    )

    return _swiftclient


# TODO(lakshmiS) See if we can cache this.
# Cached members will be equal to # of accounts at max.
def get_swiftclient_st(storageurl):
    service_type = 'object-store'
    _get_session()

    os_options = {
        'service_type': service_type,
        'region_name': CONF.service_credentials.os_region_name,
        'endpoint_type': CONF.service_credentials.os_endpoint_type,
    }
    swift_client = swiftclient.client.Connection(
        auth_version='2',
        user=CONF.service_credentials.username,
        key=CONF.service_credentials.password,
        authurl=CONF.service_credentials.auth_url,
        tenant_name=CONF.service_credentials.tenant_name,
        os_options=os_options,
        cacert=CONF.service_credentials.cafile,
        insecure=CONF.service_credentials.insecure,
        preauthurl=storageurl,
    )
    return swift_client


def get_keystoneclient():
    session = _get_session()

    return ks_client.Client(
        session=session,
        region_name=CONF.service_credentials.os_region_name)


def get_ironicclient():
    session = _get_session()
    try:
        return ironic_client.get_client(
            '1',
            session=session,
            os_region_name=CONF.service_credentials.os_region_name,
            os_endpoint_type=CONF.service_credentials.os_endpoint_type,
            os_ironic_api_version=IRONIC_API_VERSION
        )
    except ironic_exceptions.AmbiguousAuthSystem:
        raise keystone_exceptions.EndpointNotFound()



# Copyright 2016 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class LogPublisher(object):

    def publish(self, item):
        LOG.info(
            'Published notification: %(event_type)s %(doc_id)s %(doc)s' % {
                'event_type': item.event_type,
                'doc_id': item.doc_id,
                'doc': getattr(item, 'doc', None)})



# Copyright (c) 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.swift import get_swift_containers
from searchlight.elasticsearch.plugins.swift import serialize_swift_container
from searchlight.elasticsearch.plugins.swift import swift_notification_handler


class ContainerIndex(base.IndexBase):
    NotificationHandlerCls = swift_notification_handler.SwiftContainerHandler

    def get_notification_handler(self):
        """Override because the container handler needs a handle to object
        indexer for cascade delete of container to objects.
        """
        return self.NotificationHandlerCls(
            self.index_helper,
            self.options,
            object_helper=self.child_plugins[0].index_helper)

    # swift_owner_headers
    ADMIN_ONLY_FIELDS = ['x-container-write',
                         'x-container-sync-key',
                         'x-container-sync-to',
                         'x-container-meta-temp-url-key',
                         'x-container-meta-temp-url-key-2']

    @classmethod
    def parent_plugin_type(cls):
        return resource_types.SWIFT_ACCOUNT

    @classmethod
    def get_document_type(cls):
        return resource_types.SWIFT_CONTAINER

    def get_mapping(self):
        return {
            'dynamic': True,
            "_source": {
                "excludes": ["x-container-read"]
            },
            'properties': {
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    },
                },
                'account': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    },
                },
                'account_id': {'type': 'string', 'index': 'not_analyzed'},
                'x-container-read': {
                    'type': 'string', 'index': 'not_analyzed',
                    'store': False
                },

                # TODO(lakshmiS): Removing following field(s) since account
                # notifications don't include it for subsequent updates.
                # Enable when it is included in future notifications.
                # 'x-container-object-count': {'type': 'long'},
                # 'x-container-bytes-used': {'type': 'long'},

                'created_at': {'type': 'date'},
                'updated_at': {'type': 'date'}
            },
            "_parent": {
                "type": self.parent_plugin_type()
            }
        }

    @property
    def admin_only_fields(self):
        from_conf = super(ContainerIndex, self).admin_only_fields
        return ContainerIndex.ADMIN_ONLY_FIELDS + from_conf

    @classmethod
    def is_plugin_enabled_by_default(cls):
        return False

    @property
    def allow_admin_ignore_rbac(self):
        return False

    @property
    def resource_allowed_policy_target(self):
        return None

    @property
    def service_type(self):
        return None

    def _get_rbac_field_filters(self, request_context):
        tenant_member = request_context.tenant + ":*"
        single_user = request_context.tenant + ":" + request_context.user
        account_id = \
            self.parent_plugin.options.reseller_prefix + request_context.owner

        return [
            {'term': {'account_id': account_id}},
            {'terms': {'x-container-read': [tenant_member, single_user]}}
        ]

    def get_parent_id_field(self):
        return 'account_id'

    @property
    def routing_field(self):
        return "account_id"

    def get_objects(self):
        return get_swift_containers()

    def serialize(self, obj):
        return serialize_swift_container(obj)
